from typing import Any, Dict, List, Optional, Union

from deltacat.storage.model.partition import PartitionScheme
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage.model.sort_key import SortScheme
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace, NamespaceProperties
from deltacat.storage.model.schema import Schema
from deltacat.storage.model.table import TableProperties
from deltacat.storage.model.types import (
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    StreamFormat,
)
from deltacat.types.media import ContentType
from deltacat.types.tables import TableWriteMode


# catalog functions
def initialize(*args, **kwargs) -> Optional[Any]:
    """
    Initializes the data catalog with the given arguments.

    Will return an object containing any state needed for the operation of the catalog. For example,
    initializing an iceberg catalog will return the underlying native PyIceberg catalog.

    The return value initialize is stored in  :class:`deltacat.Catalog` as the "inner" property,
    and then passed to catalog function invocations as the kwarg "inner"
    """
    raise NotImplementedError("initialize not implemented")


# table functions
def write_to_table(
    data: Union[LocalTable, LocalDataset, DistributedDataset],
    table: str,
    *args,
    namespace: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    **kwargs,
) -> None:
    """Write data to a DeltaCat table.

    Args:
        data: Data to write to the table. Can be a LocalTable, LocalDataset, or DistributedDataset.
        table: Name of the table to write to.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        mode: Write mode to use when writing to the table.
        content_type: Content type of the data being written.

    Returns:
        None
    """
    raise NotImplementedError("write_to_table not implemented")


def read_table(
    table: str, *args, namespace: Optional[str] = None, **kwargs
) -> DistributedDataset:
    """Read data from a DeltaCat table.

    Args:
        table: Name of the table to read from.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        A Deltacat DistributedDataset containing the table data.
    """
    raise NotImplementedError("read_table not implemented")


def alter_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[Dict[str, Any]] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    properties: Optional[TableProperties] = None,
    **kwargs,
) -> None:
    """Alter deltacat table/table_version definition.

    Modifies various aspects of a table's metadata including lifecycle state,
    schema, partitioning, sort keys, description, and properties.

    Args:
        table: Name of the table to alter.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        lifecycle_state: New lifecycle state for the table.
        schema_updates: Map of schema updates to apply.
        partition_updates: Map of partition scheme updates to apply.
        sort_keys: New sort keys scheme.
        description: New description for the table.
        properties: New table properties.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not already exist.
    """
    raise NotImplementedError("alter_table not implemented")


def create_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    version: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = LifecycleState.ACTIVE,
    schema: Optional[Schema] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    namespace_properties: Optional[NamespaceProperties] = None,
    content_types: Optional[List[ContentType]] = None,
    fail_if_exists: bool = True,
    **kwargs,
) -> TableDefinition:
    """Create an empty table in the catalog.

    If a namespace isn't provided, the table will be created within the default deltacat namespace.
    Additionally if the provided namespace does not exist, it will be created for you.

    Args:
        name: Name of the table to create.
        namespace: Optional namespace for the table. Uses default namespace if not specified.
        version: Optional version identifier for the table.
        lifecycle_state: Lifecycle state of the new table. Defaults to ACTIVE.
        schema: Schema definition for the table.
        partition_scheme: Optional partitioning scheme for the table.
        sort_keys: Optional sort keys for the table.
        description: Optional description of the table.
        table_properties: Optional properties for the table.
        namespace_properties: Optional properties for the namespace if it needs to be created.
        content_types: Optional list of allowed content types for the table.
        fail_if_exists: If True, raises an error if table already exists. If False, returns existing table.

    Returns:
        TableDefinition object for the created or existing table.

    Raises:
        TableAlreadyExistsError: If the table already exists and fail_if_exists is True.
        NamespaceNotFoundError: If the provided namespace does not exist.
    """
    raise NotImplementedError("create_table not implemented")


def drop_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    purge: bool = False,
    **kwargs,
) -> None:
    """Drop a table from the catalog and optionally purges underlying data.

    Args:
        name: Name of the table to drop.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to drop.
        purge: If True, permanently delete the table data. If False, only remove from catalog.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not exist.
    """
    raise NotImplementedError("drop_table not implemented")


def refresh_table(table: str, *args, namespace: Optional[str] = None, **kwargs) -> None:
    """Refresh metadata cached on the Ray cluster for the given table.

    Args:
        table: Name of the table to refresh.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        None
    """
    raise NotImplementedError("refresh_table not implemented")


def list_tables(
    *args, namespace: Optional[str] = None, **kwargs
) -> ListResult[TableDefinition]:
    """List a page of table definitions.

    Args:
        namespace: Optional namespace to list tables from. Uses default namespace if not specified.

    Returns:
        ListResult containing TableDefinition objects for tables in the namespace.
    """
    raise NotImplementedError("list_tables not implemented")


def get_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    **kwargs,
) -> Optional[TableDefinition]:
    """Get table definition metadata.

    Args:
        name: Name of the table to retrieve.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to retrieve.
            If not specified, the latest version is used.
        stream_format: Optional stream format to retrieve. Uses the default Deltacat stream
            format if not specified.

    Returns:
        Deltacat TableDefinition if the table exists, None otherwise.

    Raises:
        TableVersionNotFoundError: If the table version does not exist.
        StreamNotFoundError: If the stream does not exist.
    """
    raise NotImplementedError("get_table not implemented")


def truncate_table(
    table: str, *args, namespace: Optional[str] = None, **kwargs
) -> None:
    """Truncate table data.

    Args:
        table: Name of the table to truncate.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        None
    """
    raise NotImplementedError("truncate_table not implemented")


def rename_table(
    table: str, new_name: str, *args, namespace: Optional[str] = None, **kwargs
) -> None:
    """Rename an existing table.

    Args:
        table: Current name of the table.
        new_name: New name for the table.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not exist.
    """
    raise NotImplementedError("rename_table not implemented")


def table_exists(table: str, *args, namespace: Optional[str] = None, **kwargs) -> bool:
    """Check if a table exists in the catalog.

    Args:
        table: Name of the table to check.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        True if the table exists, False otherwise.
    """
    raise NotImplementedError("table_exists not implemented")


# namespace functions
def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """List a page of table namespaces.

    Args:
        catalog: Catalog properties instance.

    Returns:
        ListResult containing Namespace objects.
    """
    raise NotImplementedError("list_namespaces not implemented")


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """Get metadata for a specific table namespace.

    Args:
        namespace: Name of the namespace to retrieve.

    Returns:
        Namespace object if the namespace exists, None otherwise.
    """
    raise NotImplementedError("get_namespace not implemented")


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """Check if a namespace exists.

    Args:
        namespace: Name of the namespace to check.

    Returns:
        True if the namespace exists, False otherwise.
    """
    raise NotImplementedError("namespace_exists not implemented")


def create_namespace(
    namespace: str,
    properties: Optional[NamespaceProperties] = None,
    *args,
    **kwargs,
) -> Namespace:
    """Create a new namespace.

    Args:
        namespace: Name of the namespace to create.
        properties: Optional properties for the namespace.

    Returns:
        Created Namespace object.

    Raises:
        NamespaceAlreadyExistsError: If the namespace already exists.
    """
    raise NotImplementedError("create_namespace not implemented")


def alter_namespace(
    namespace: str,
    *args,
    properties: Optional[NamespaceProperties] = None,
    new_namespace: Optional[str] = None,
    **kwargs,
) -> None:
    """Alter a namespace definition.

    Args:
        namespace: Name of the namespace to alter.
        properties: Optional new properties for the namespace.
        new_namespace: Optional new name for the namespace.

    Returns:
        None
    """
    raise NotImplementedError("alter_namespace not implemented")


def drop_namespace(namespace: str, *args, purge: bool = False, **kwargs) -> None:
    """Drop a namespace and all of its tables from the catalog.

    Args:
        namespace: Name of the namespace to drop.
        purge: If True, permanently delete all tables in the namespace.
            If False, only remove from catalog.

    Returns:
        None
    """
    raise NotImplementedError("drop_namespace not implemented")


def default_namespace(*args, **kwargs) -> str:
    """Return the default namespace for the catalog.

    Returns:
        String name of the default namespace.
    """
    raise NotImplementedError("default_namespace not implemented")



from typing import Optional

import pyarrow as pa
from pyarrow import Field as PaField
from daft import Schema as DaftSchema, DataType
from daft.daft import (
    PartitionField as DaftPartitionField,
    PartitionTransform as DaftTransform,
)
from daft.logical.schema import Field as DaftField
from daft.io.scan import make_partition_field

from deltacat.storage.model.schema import Schema
from deltacat.storage.model.interop import ModelMapper
from deltacat.storage.model.partition import PartitionKey
from deltacat.storage.model.transform import (
    BucketingStrategy,
    Transform,
    BucketTransform,
    HourTransform,
    DayTransform,
    MonthTransform,
    YearTransform,
    IdentityTransform,
    TruncateTransform,
)


class DaftFieldMapper(ModelMapper[DaftField, PaField]):
    @staticmethod
    def map(
        obj: Optional[DaftField],
        **kwargs,
    ) -> Optional[PaField]:
        """Convert Daft Field to PyArrow Field.

        Args:
            obj: The Daft Field to convert
            **kwargs: Additional arguments

        Returns:
            Converted PyArrow Field object
        """
        if obj is None:
            return None

        return pa.field(
            name=obj.name,
            type=obj.dtype.to_arrow_dtype(),
        )

    @staticmethod
    def unmap(
        obj: Optional[PaField],
        **kwargs,
    ) -> Optional[DaftField]:
        """Convert PyArrow Field to Daft Field.

        Args:
            obj: The PyArrow Field to convert
            **kwargs: Additional arguments

        Returns:
            Converted Daft Field object
        """
        if obj is None:
            return None

        return DaftField.create(
            name=obj.name,
            dtype=DataType.from_arrow_type(obj.type),  # type: ignore
        )


class DaftTransformMapper(ModelMapper[DaftTransform, Transform]):
    @staticmethod
    def map(
        obj: Optional[DaftTransform],
        **kwargs,
    ) -> Optional[Transform]:
        """Convert DaftTransform to DeltaCAT Transform.

        Args:
            obj: The DaftTransform to convert
            **kwargs: Additional arguments

        Returns:
            Converted Transform object
        """

        # daft.PartitionTransform doesn't have a Python interface for accessing its attributes,
        # thus conversion is not possible.
        # TODO: request Daft to expose Python friendly interface for daft.PartitionTransform
        raise NotImplementedError(
            "Converting transform from Daft to DeltaCAT is not supported"
        )

    @staticmethod
    def unmap(
        obj: Optional[Transform],
        **kwargs,
    ) -> Optional[DaftTransform]:
        """Convert DeltaCAT Transform to DaftTransform.

        Args:
            obj: The Transform to convert
            **kwargs: Additional arguments

        Returns:
            Converted DaftTransform object
        """
        if obj is None:
            return None

        # Map DeltaCAT transforms to Daft transforms using isinstance

        if isinstance(obj, IdentityTransform):
            return DaftTransform.identity()
        elif isinstance(obj, HourTransform):
            return DaftTransform.hour()
        elif isinstance(obj, DayTransform):
            return DaftTransform.day()
        elif isinstance(obj, MonthTransform):
            return DaftTransform.month()
        elif isinstance(obj, YearTransform):
            return DaftTransform.year()
        elif isinstance(obj, BucketTransform):
            if obj.parameters.bucketing_strategy == BucketingStrategy.ICEBERG:
                return DaftTransform.iceberg_bucket(obj.parameters.num_buckets)
            else:
                raise ValueError(
                    f"Unsupported Bucketing Strategy: {obj.parameters.bucketing_strategy}"
                )
        elif isinstance(obj, TruncateTransform):
            return DaftTransform.iceberg_truncate(obj.parameters.width)

        raise ValueError(f"Unsupported Transform: {obj}")


class DaftPartitionKeyMapper(ModelMapper[DaftPartitionField, PartitionKey]):
    @staticmethod
    def map(
        obj: Optional[DaftPartitionField],
        schema: Optional[DaftSchema] = None,
        **kwargs,
    ) -> Optional[PartitionKey]:
        """Convert DaftPartitionField to PartitionKey.

        Args:
            obj: The DaftPartitionField to convert
            schema: The Daft schema containing field information
            **kwargs: Additional arguments

        Returns:
            Converted PartitionKey object
        """
        # Daft PartitionField only exposes 1 attribute `field` which is not enough
        # to convert to DeltaCAT PartitionKey
        # TODO: request Daft to expose more Python friendly interface for PartitionField
        raise NotImplementedError(
            f"Converting Daft PartitionField to DeltaCAT PartitionKey is not supported"
        )

    @staticmethod
    def unmap(
        obj: Optional[PartitionKey],
        schema: Optional[Schema] = None,
        **kwargs,
    ) -> Optional[DaftPartitionField]:
        """Convert PartitionKey to DaftPartitionField.

        Args:
            obj: The DeltaCAT PartitionKey to convert
            schema: The Schema containing field information
            **kwargs: Additional arguments

        Returns:
            Converted DaftPartitionField object
        """
        if obj is None:
            return None
        if obj.name is None:
            raise ValueError("Name is required for PartitionKey conversion")
        if not schema:
            raise ValueError("Schema is required for PartitionKey conversion")
        if len(obj.key) < 1:
            raise ValueError(
                f"At least 1 PartitionKey FieldLocator is expected, instead got {len(obj.key)}. FieldLocators: {obj.key}."
            )

        # Get the source field from schema - FieldLocator in PartitionKey.key points to the source field of partition field
        dc_source_field = schema.field(obj.key[0]).arrow
        daft_source_field = DaftFieldMapper.unmap(obj=dc_source_field)
        # Convert transform if present
        daft_transform = DaftTransformMapper.unmap(obj.transform)
        daft_partition_field = DaftPartitionKeyMapper.get_daft_partition_field(
            partition_field_name=obj.name,
            daft_source_field=daft_source_field,
            dc_transform=obj.transform,
        )

        # Create DaftPartitionField
        return make_partition_field(
            field=daft_partition_field,
            source_field=daft_source_field,
            transform=daft_transform,
        )

    @staticmethod
    def get_daft_partition_field(
        partition_field_name: str,
        daft_source_field: Optional[DaftField],
        # TODO: replace DeltaCAT transform with Daft Transform for uniformality
        # We cannot use Daft Transform here because Daft Transform doesn't have a Python interface for us to
        # access its attributes.
        # TODO: request Daft to provide a more python friendly interface for Daft Tranform
        dc_transform: Optional[Transform],
    ) -> DaftField:
        """Generate Daft Partition Field given partition field name, source field and transform.
        Partition field type is inferred using source field type and transform.

        Args:
            partition_field_name (str): the specified result field name
            daft_source_field (DaftField): the source field of the partition field
            daft_transform (DaftTransform): transform applied on the source field to create partition field

        Returns:
            DaftField: Daft Field representing the partition field
        """
        if daft_source_field is None:
            raise ValueError("Source field is required for PartitionField conversion")
        if dc_transform is None:
            raise ValueError("Transform is required for PartitionField conversion")

        result_type = None
        # Below type conversion logic references Daft - Iceberg conversion logic:
        # https://github.com/Eventual-Inc/Daft/blob/7f2e9b5fb50fdfe858be17572f132b37dd6e5ab2/daft/iceberg/iceberg_scan.py#L61-L85
        if isinstance(dc_transform, IdentityTransform):
            result_type = daft_source_field.dtype
        elif isinstance(dc_transform, YearTransform):
            result_type = DataType.int32()
        elif isinstance(dc_transform, MonthTransform):
            result_type = DataType.int32()
        elif isinstance(dc_transform, DayTransform):
            result_type = DataType.int32()
        elif isinstance(dc_transform, HourTransform):
            result_type = DataType.int32()
        elif isinstance(dc_transform, BucketTransform):
            result_type = DataType.int32()
        elif isinstance(dc_transform, TruncateTransform):
            result_type = daft_source_field.dtype
        else:
            raise ValueError(f"Unsupported transform: {dc_transform}")

        return DaftField.create(
            name=partition_field_name,
            dtype=result_type,
        )



from typing import Any, Dict, List, Optional, Union

from deltacat.catalog.model.catalog import get_catalog
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage.model.partition import PartitionScheme
from deltacat.storage.model.sort_key import SortScheme
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace, NamespaceProperties
from deltacat.storage.model.schema import Schema
from deltacat.storage.model.table import TableProperties
from deltacat.storage.model.types import (
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    StreamFormat,
)
from deltacat.types.media import ContentType
from deltacat.types.tables import TableWriteMode


# table functions
def write_to_table(
    data: Union[LocalTable, LocalDataset, DistributedDataset],
    table: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    **kwargs,
) -> None:
    """Write local or distributed data to a table. Raises an error if the
    table does not exist and the table write mode is not CREATE or AUTO.

    When creating a table, all `create_table` parameters may be optionally
    specified as additional keyword arguments. When appending to, or replacing,
    an existing table, all `alter_table` parameters may be optionally specified
    as additional keyword arguments."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.write_to_table(
        data,
        table,
        *args,
        namespace=namespace,
        mode=mode,
        content_type=content_type,
        inner=catalog_obj.inner,
        **kwargs,
    )


def read_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> DistributedDataset:
    """Read a table into a distributed dataset."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.read_table(
        table,
        *args,
        namespace=namespace,
        inner=catalog_obj.inner,
        **kwargs,
    )


def alter_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[Dict[str, Any]] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    properties: Optional[TableProperties] = None,
    **kwargs,
) -> None:
    """Alter table definition."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.alter_table(
        table,
        *args,
        namespace=namespace,
        lifecycle_state=lifecycle_state,
        schema_updates=schema_updates,
        partition_updates=partition_updates,
        sort_keys=sort_keys,
        description=description,
        properties=properties,
        inner=catalog_obj.inner,
        **kwargs,
    )


def create_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    version: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = LifecycleState.ACTIVE,
    schema: Optional[Schema] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    namespace_properties: Optional[NamespaceProperties] = None,
    content_types: Optional[List[ContentType]] = None,
    fail_if_exists: bool = True,
    **kwargs,
) -> TableDefinition:
    """Create an empty table. Raises an error if the table already exists and
    `fail_if_exists` is True (default behavior)."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.create_table(
        name,
        *args,
        namespace=namespace,
        version=version,
        lifecycle_state=lifecycle_state,
        schema=schema,
        partition_scheme=partition_scheme,
        sort_keys=sort_keys,
        description=description,
        table_properties=table_properties,
        namespace_properties=namespace_properties,
        content_types=content_types,
        fail_if_exists=fail_if_exists,
        inner=catalog_obj.inner,
        **kwargs,
    )


def drop_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    table_version: Optional[str] = None,
    purge: bool = False,
    **kwargs,
) -> None:
    """Drop a table from the catalog and optionally purge it. Raises an error
    if the table does not exist."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.drop_table(
        name,
        *args,
        namespace=namespace,
        table_version=table_version,
        purge=purge,
        inner=catalog_obj.inner,
        **kwargs,
    )


def refresh_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Refresh metadata cached on the Ray cluster for the given table."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.refresh_table(
        table,
        *args,
        namespace=namespace,
        inner=catalog_obj.inner,
        **kwargs,
    )


def list_tables(
    *args, namespace: Optional[str] = None, catalog: Optional[str] = None, **kwargs
) -> ListResult[TableDefinition]:
    """List a page of table definitions. Raises an error if the given namespace
    does not exist."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.list_tables(
        *args,
        namespace=namespace,
        inner=catalog_obj.inner,
        **kwargs,
    )


def get_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    **kwargs,
) -> Optional[TableDefinition]:
    """Get table definition metadata. Returns None if the given table does not
    exist."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.get_table(
        name,
        *args,
        namespace=namespace,
        table_version=table_version,
        stream_format=stream_format,
        inner=catalog_obj.inner,
        **kwargs,
    )


def truncate_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Truncate table data. Raises an error if the table does not exist."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.truncate_table(
        table,
        *args,
        namespace=namespace,
        inner=catalog_obj.inner,
        **kwargs,
    )


def rename_table(
    table: str,
    new_name: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Rename a table."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.rename_table(
        table,
        new_name,
        *args,
        namespace=namespace,
        inner=catalog_obj.inner,
        **kwargs,
    )


def table_exists(
    table: str,
    *args,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> bool:
    """Returns True if the given table exists, False if not."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.table_exists(
        table,
        *args,
        namespace=namespace,
        inner=catalog_obj.inner,
        **kwargs,
    )


# namespace functions
def list_namespaces(
    *args, catalog: Optional[str] = None, **kwargs
) -> ListResult[Namespace]:
    """List a page of table namespaces."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.list_namespaces(
        *args,
        inner=catalog_obj.inner,
        **kwargs,
    )


def get_namespace(
    namespace: str,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Namespace]:
    """Get table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.get_namespace(
        namespace,
        *args,
        inner=catalog_obj.inner,
        **kwargs,
    )


def namespace_exists(
    namespace: str,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> bool:
    """Returns True if the given table namespace exists, False if not."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.namespace_exists(
        namespace,
        *args,
        inner=catalog_obj.inner,
        **kwargs,
    )


def create_namespace(
    namespace: str,
    properties: Optional[NamespaceProperties] = None,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> Namespace:
    """Creates a table namespace with the given name and properties. Returns
    the created namespace. Raises an error if the namespace already exists."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.create_namespace(
        namespace,
        *args,
        properties=properties,
        inner=catalog_obj.inner,
        **kwargs,
    )


def alter_namespace(
    namespace: str,
    *args,
    catalog: Optional[str] = None,
    properties: Optional[NamespaceProperties] = None,
    new_namespace: Optional[str] = None,
    **kwargs,
) -> None:
    """Alter table namespace definition."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.alter_namespace(
        namespace,
        *args,
        properties=properties,
        new_namespace=new_namespace,
        inner=catalog_obj.inner,
        **kwargs,
    )


def drop_namespace(
    namespace: str, *args, catalog: Optional[str] = None, purge: bool = False, **kwargs
) -> None:
    """Drop the given namespace and all of its tables from the catalog,
    optionally purging them."""
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.drop_namespace(
        namespace,
        *args,
        purge=purge,
        inner=catalog_obj.inner,
        **kwargs,
    )


def default_namespace(*args, catalog: Optional[str] = None, **kwargs) -> str:
    """Returns the default namespace for the catalog."""
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.default_namespace(*args, inner=catalog_obj.inner, **kwargs)



from __future__ import annotations
from enum import Enum
from typing import Callable
import logging

import tenacity

from pyarrow.lib import ArrowException, ArrowInvalid, ArrowCapacityError

import botocore
from botocore.exceptions import BotoCoreError

import ray
from ray.exceptions import (
    RayError,
    RayTaskError,
    RuntimeEnvSetupError,
    WorkerCrashedError,
    NodeDiedError,
    OutOfMemoryError,
)

from daft.exceptions import DaftTransientError, DaftCoreException

import deltacat as dc
from deltacat import logs
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

DELTACAT_STORAGE_PARAM = "deltacat_storage"
DELTACAT_STORAGE_KWARGS_PARAM = "deltacat_storage_kwargs"


class DeltaCatErrorNames(str, Enum):

    DEPENDENCY_RAY_ERROR = "DependencyRayError"
    DEPENDENCY_RAY_WORKER_DIED_ERROR = "DependencyRayWorkerDiedError"
    DEPENDENCY_RAY_OUT_OF_MEMORY_ERROR = "DependencyRayOOMError"
    DEPENDENCY_RAY_RUNTIME_SETUP_ERROR = "DependencyRayRuntimeSetupError"
    DEPENDENCY_BOTOCORE_ERROR = "DependencyBotocoreError"
    DEPENDENCY_BOTOCORE_CONNECTION_ERROR = "DependencyBotocoreConnectionError"
    DEPENDENCY_BOTOCORE_CREDENTIAL_ERROR = "DependencyBotocoreCredentialError"
    DEPENDENCY_BOTOCORE_TIMEOUT_ERROR = "DependencyBotocoreTimeoutError"
    NON_RETRYABLE_DOWNLOAD_TABLE_ERROR = "NonRetryableDownloadTableError"
    NON_RETRYABLE_DOWNLOAD_FILE_ERROR = "NonRetryableDownloadFileError"
    NON_RETRYABLE_UPLOAD_TABLE_ERROR = "NonRetryableUploadTableError"
    NON_RETRYABLE_UPLOAD_FILE_ERROR = "NonRetryableUploadFileError"
    DEPENDENCY_PYARROW_ERROR = "DependencyPyarrowError"
    DEPENDENCY_PYARROW_INVALID_ERROR = "DependencyPyarrowInvalidError"
    DEPENDENCY_PYARROW_CAPACITY_ERROR = "DependencyPyarrowCapacityError"
    PYMEMCACHED_PUT_OBJECT_ERROR = "PymemcachedPutObjectError"
    DEPENDENCY_DAFT_ERROR = "DependencyDaftError"

    GENERAL_THROTTLING_ERROR = "GeneralThrottlingError"
    RETRYABLE_UPLOAD_TABLE_ERROR = "RetryableUploadTableError"
    RETRYABLE_UPLOAD_FILE_ERROR = "RetryableUploadFileError"
    RETRYABLE_DOWNLOAD_FILE_ERROR = "RetryableDownloadFileError"
    RETRYABLE_DOWNLOAD_TABLE_ERROR = "RetryableDownloadTableError"
    RETRYABLE_TIMEOUT_ERROR = "RetryableTimeoutError"
    DEPENDENCY_DAFT_TRANSIENT_ERROR = "DependencyDaftTransientError"

    VALIDATION_ERROR = "ValidationError"
    CONTENT_TYPE_VALIDATION_ERROR = "ContentTypeValidationError"

    DELTACAT_SYSTEM_ERROR = "DeltaCatSystemError"
    DELTACAT_TRANSIENT_ERROR = "DeltaCatTransientError"
    UNCLASSIFIED_DELTACAT_ERROR = "UnclassifiedDeltaCatError"
    UNRECOGNIZED_RAY_TASK_ERROR = "UnrecognizedRayTaskError"

    NAMESPACE_NOT_FOUND_ERROR = "NamespaceNotFoundError"
    TABLE_NOT_FOUND_ERROR = "TableNotFoundError"
    TABLE_VERSION_NOT_FOUND_ERROR = "TableVersionNotFoundError"
    STREAM_NOT_FOUND_ERROR = "StreamNotFoundError"
    DELTA_NOT_FOUND_ERROR = "DeltaNotFoundError"
    TABLE_ALREADY_EXISTS_ERROR = "TableAlreadyExistsError"
    NAMESPACE_ALREADY_EXISTS_ERROR = "NamespaceAlreadyExistsError"


class DeltaCatError(Exception):
    def __init__(self, *args, **kwargs):
        task_id, node_ip = self._get_ray_task_id_and_node_ip()
        self.task_id = task_id
        self.node_ip = node_ip
        super().__init__(*args, **kwargs)

    def _get_ray_task_id_and_node_ip(self):
        task_id = get_current_ray_task_id()
        node_ip = ray.util.get_node_ip_address()
        return task_id, node_ip


class NonRetryableError(DeltaCatError):
    is_retryable = False


class RetryableError(DeltaCatError):
    is_retryable = True


class ValidationError(NonRetryableError):
    error_name = DeltaCatErrorNames.VALIDATION_ERROR.value


class UnclassifiedDeltaCatError(NonRetryableError):
    error_name = DeltaCatErrorNames.UNCLASSIFIED_DELTACAT_ERROR.value


class DependencyRayError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_ERROR.value


class DeltaCatTransientError(RetryableError):
    error_name = DeltaCatErrorNames.DELTACAT_TRANSIENT_ERROR.value


class DependencyDaftError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_DAFT_ERROR.value


class DependencyRayWorkerDiedError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_WORKER_DIED_ERROR.value


class DependencyRayOutOfMemoryError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_OUT_OF_MEMORY_ERROR.value


class DependencyRayRuntimeSetupError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_RUNTIME_SETUP_ERROR.value


class DependencyPyarrowError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_PYARROW_ERROR.value


class DependencyPyarrowInvalidError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_PYARROW_INVALID_ERROR.value


class DependencyPyarrowCapacityError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_PYARROW_CAPACITY_ERROR.value


class PymemcachedPutObjectError(RetryableError):
    error_name = DeltaCatErrorNames.PYMEMCACHED_PUT_OBJECT_ERROR.value


class ContentTypeValidationError(NonRetryableError):
    error_name = DeltaCatErrorNames.CONTENT_TYPE_VALIDATION_ERROR.value


class DependencyBotocoreError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_ERROR.value


class DependencyBotocoreConnectionError(DeltaCatTransientError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_CONNECTION_ERROR.value


class DependencyBotocoreCredentialError(DeltaCatTransientError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_CREDENTIAL_ERROR.value


class DependencyBotocoreTimeoutError(DeltaCatTransientError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_TIMEOUT_ERROR.value


class NonRetryableDownloadFileError(NonRetryableError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_DOWNLOAD_FILE_ERROR.value


class NonRetryableDownloadTableError(NonRetryableDownloadFileError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_DOWNLOAD_TABLE_ERROR.value


class NonRetryableUploadFileError(NonRetryableError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_UPLOAD_FILE_ERROR.value


class NonRetryableUploadTableError(NonRetryableUploadFileError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_UPLOAD_TABLE_ERROR.value


class GeneralThrottlingError(RetryableError):
    error_name = DeltaCatErrorNames.GENERAL_THROTTLING_ERROR.value


class RetryableUploadFileError(RetryableError):
    error_name = DeltaCatErrorNames.RETRYABLE_UPLOAD_FILE_ERROR.value


class RetryableUploadTableError(RetryableUploadFileError):
    error_name = DeltaCatErrorNames.RETRYABLE_UPLOAD_TABLE_ERROR.value


class RetryableDownloadFileError(RetryableError):
    error_name = DeltaCatErrorNames.RETRYABLE_DOWNLOAD_FILE_ERROR.value


class RetryableDownloadTableError(RetryableDownloadFileError):
    error_name = DeltaCatErrorNames.RETRYABLE_DOWNLOAD_TABLE_ERROR.value


class RetryableTimeoutError(RetryableError):
    error_name = DeltaCatErrorNames.RETRYABLE_TIMEOUT_ERROR.value


class DependencyDaftTransientError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_DAFT_TRANSIENT_ERROR.value


class DeltaCatSystemError(NonRetryableError):
    error_name = DeltaCatErrorNames.DELTACAT_SYSTEM_ERROR.value


class UnrecognizedRayTaskError(NonRetryableError):
    error_name = DeltaCatErrorNames.UNRECOGNIZED_RAY_TASK_ERROR.value


class NamespaceNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.NAMESPACE_NOT_FOUND_ERROR.value


class TableNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_NOT_FOUND_ERROR.value


class TableVersionNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_VERSION_NOT_FOUND_ERROR.value


class StreamNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.STREAM_NOT_FOUND_ERROR.value


class DeltaNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.DELTA_NOT_FOUND_ERROR.value


class TableAlreadyExistsError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_ALREADY_EXISTS_ERROR.value


class NamespaceAlreadyExistsError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_ALREADY_EXISTS_ERROR.value


def categorize_errors(func: Callable):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            deltacat_storage = None
            deltacat_storage_kwargs = {}
            all_args = args
            if kwargs:
                deltacat_storage = kwargs.get(DELTACAT_STORAGE_PARAM)
                deltacat_storage_kwargs = kwargs.get(DELTACAT_STORAGE_KWARGS_PARAM, {})
                all_args = all_args + tuple(kwargs.values())

            if not deltacat_storage and all_args:
                for arg in all_args:
                    if (
                        isinstance(arg, dict)
                        and arg.get(DELTACAT_STORAGE_PARAM) is not None
                    ):
                        deltacat_storage = arg.get(DELTACAT_STORAGE_PARAM)
                        deltacat_storage_kwargs = arg.get(
                            DELTACAT_STORAGE_KWARGS_PARAM, {}
                        )
                        break

            categorize_deltacat_exception(e, deltacat_storage, deltacat_storage_kwargs)

    return wrapper


def categorize_deltacat_exception(
    e: BaseException,
    deltacat_storage: dc.storage.interface = None,
    deltacat_storage_kwargs: dict = None,
):
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}

    if isinstance(e, DeltaCatError):
        raise e
    elif deltacat_storage and deltacat_storage.can_categorize(
        e, **deltacat_storage_kwargs
    ):
        deltacat_storage.raise_categorized_error(e, **deltacat_storage_kwargs)
    elif isinstance(e, RayError):
        _categorize_ray_error(e)
    elif isinstance(e, tenacity.RetryError):
        _categorize_tenacity_error(e)
    elif isinstance(e, ArrowException):
        _categorize_dependency_pyarrow_error(e)
    elif isinstance(e, AssertionError):
        _categorize_assertion_error(e)
    elif isinstance(e, DaftCoreException):
        _categorize_daft_error(e)
    elif isinstance(e, BotoCoreError):
        _categorize_botocore_error(e)
    else:
        _categorize_all_remaining_errors(e)

    logger.error(f"Error categorization failed for {e}.", exc_info=True)
    raise UnclassifiedDeltaCatError(
        "Error could not categorized into DeltaCat error"
    ) from e


def _categorize_ray_error(e: RayError):
    if isinstance(e, RuntimeEnvSetupError):
        raise DependencyRayRuntimeSetupError("Ray failed to setup runtime env.") from e
    elif isinstance(e, WorkerCrashedError) or isinstance(e, NodeDiedError):
        raise DependencyRayWorkerDiedError("Ray worker died unexpectedly.") from e
    elif isinstance(e, OutOfMemoryError):
        raise DependencyRayOutOfMemoryError("Ray worker Out Of Memory.") from e
    elif isinstance(e, RayTaskError):
        if e.cause is not None and isinstance(e.cause, Exception):
            categorize_deltacat_exception(e.cause)
        else:
            raise UnrecognizedRayTaskError(
                "Unrecognized underlying error detected in a Ray task."
            ) from e
    else:
        raise DependencyRayError("Dependency Ray error occurred.") from e


def _categorize_tenacity_error(e: tenacity.RetryError):
    if e.__cause__ is not None and isinstance(e.__cause__, Exception):
        categorize_deltacat_exception(e.__cause__)
    else:
        raise RetryableError("Unrecognized retryable error occurred.") from e


def _categorize_dependency_pyarrow_error(e: ArrowException):
    if isinstance(e, ArrowInvalid):
        raise DependencyPyarrowInvalidError(
            f"Pyarrow Invalid error occurred. {e}"
        ) from e
    elif isinstance(e, ArrowCapacityError):
        raise DependencyPyarrowCapacityError("Pyarrow Capacity error occurred.") from e
    else:
        raise DependencyPyarrowError("Pyarrow error occurred.") from e


def _categorize_assertion_error(e: BaseException):
    raise ValidationError(f"One of the assertions in DeltaCAT has failed. {e}") from e


def _categorize_daft_error(e: DaftCoreException):
    if isinstance(e, DaftTransientError):
        raise DependencyDaftTransientError("Daft Transient error occurred.") from e
    elif isinstance(e, DaftCoreException):
        raise DependencyDaftError("Daft error occurred.") from e


def _categorize_botocore_error(e: BotoCoreError):
    if isinstance(e, botocore.exceptions.ConnectionError) or isinstance(
        e, botocore.exceptions.HTTPClientError
    ):
        raise DependencyBotocoreConnectionError(
            "Botocore connection error occurred."
        ) from e
    elif isinstance(e, botocore.exceptions.CredentialRetrievalError) or isinstance(
        e, botocore.exceptions.NoCredentialsError
    ):
        raise DependencyBotocoreCredentialError(
            "Botocore credential retrieval failed"
        ) from e
    elif isinstance(e, botocore.exceptions.ReadTimeoutError) or isinstance(
        e, botocore.exceptions.ConnectTimeoutError
    ):
        raise DependencyBotocoreTimeoutError("Botocore connection timed out.") from e
    else:
        raise DependencyBotocoreError("Botocore error occurred.") from e


def _categorize_all_remaining_errors(e: BaseException):
    if isinstance(e, ConnectionError):
        raise DeltaCatTransientError("Connection error has occurred.") from e
    elif isinstance(e, TimeoutError):
        raise DeltaCatTransientError("Timeout error has occurred.") from e
    elif isinstance(e, OSError):
        raise DeltaCatTransientError("OSError occurred.") from e
    elif isinstance(e, SystemExit):
        raise DeltaCatSystemError("Unexpected System error occurred.") from e



from collections import defaultdict
import logging
from deltacat import logs
from deltacat.compute.converter.model.convert_input_files import ConvertInputFiles

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def check_data_files_sequence_number(data_files_list, equality_delete_files_list):
    # Sort by file sequence number
    data_files_list.sort(key=lambda file_tuple: file_tuple[0])
    equality_delete_files_list.sort(key=lambda file_tuple: file_tuple[0])

    equality_delete_files = []
    result_data_file = []

    # Pointer for list data_file
    data_file_pointer = 0

    # Loop through each value in equality_delete_file
    for equality_file_tuple in equality_delete_files_list:
        # Find all values in data_file that are smaller than val_equality
        valid_values = []

        # Move data_file_pointer to the first value in data_file that is smaller than val_equality
        while (
            data_file_pointer < len(data_files_list)
            and data_files_list[data_file_pointer][0] < equality_file_tuple[0]
        ):
            valid_values.append(data_files_list[data_file_pointer])
            data_file_pointer += 1
            equality_delete_files.append(equality_file_tuple)

        # Append the value from equality_delete_file and the corresponding valid values from data_file
        if valid_values:
            result_data_file.append(valid_values)

    result_equality_delete_file = append_larger_sequence_number_data_files(
        equality_delete_files
    )

    return result_equality_delete_file, result_data_file


def append_larger_sequence_number_data_files(data_files_list):
    result = []
    # Iterate over the input list
    for i in range(len(data_files_list)):
        sublist = data_files_list[i:]
        sublist_file_list = []
        for file in sublist:
            sublist_file_list.append(file)
        result.append(sublist_file_list)
    return result


def construct_iceberg_table_prefix(
    iceberg_warehouse_bucket_name, table_name, iceberg_namespace
):
    return f"{iceberg_warehouse_bucket_name}/{iceberg_namespace}/{table_name}/data"


def partition_value_record_to_partition_value_string(partition):
    # Get string representation of partition value out of Record[partition_value]
    partition_value_str = partition.__repr__().split("[", 1)[1].split("]")[0]
    return partition_value_str


def group_all_files_to_each_bucket(
    data_file_dict, equality_delete_dict, pos_delete_dict
):
    convert_input_files_for_all_buckets = []
    files_for_each_bucket_for_deletes = defaultdict(tuple)
    if equality_delete_dict:
        for partition_value, equality_delete_file_list in equality_delete_dict.items():
            (
                result_equality_delete_file,
                result_data_file,
            ) = check_data_files_sequence_number(
                data_files_list=data_file_dict[partition_value],
                equality_delete_files_list=equality_delete_dict[partition_value],
            )
            files_for_each_bucket_for_deletes[partition_value] = (
                result_data_file,
                result_equality_delete_file,
                [],
            )
            if partition_value not in data_file_dict:
                convert_input_file = ConvertInputFiles.of(
                    partition_value=partition_value,
                    applicable_data_files=result_data_file,
                    applicable_equalitu_delete_files=result_equality_delete_file,
                )
                convert_input_files_for_all_buckets.append(convert_input_file)

    for partition_value, all_data_files_for_each_bucket in data_file_dict.items():
        convert_input_file = ConvertInputFiles.of(
            partition_value=partition_value,
            all_data_files_for_dedupe=all_data_files_for_each_bucket,
        )
        if partition_value in files_for_each_bucket_for_deletes:
            convert_input_file.applicable_data_files = (
                files_for_each_bucket_for_deletes[partition_value][0]
            )
            convert_input_file.applicable_delete_files = (
                files_for_each_bucket_for_deletes[partition_value][1]
            )
        convert_input_files_for_all_buckets.append(convert_input_file)
    return convert_input_files_for_all_buckets



import pytest

from deltacat.storage import (
    Transaction,
    TransactionOperation,
    TransactionType,
    TransactionOperationType,
)
from deltacat.storage.model.metafile import (
    Metafile,
)


class TestAbsToRelative:
    @classmethod
    def setup_method(cls):
        cls.catalog_root = "/catalog/root/path"

    # Test cases for the abs_to_relative function
    def test_abs_to_relative_simple(self):
        """
        Tests the function which relativizes absolute paths (string) into relative paths (string)
        """
        catalog_root = TestAbsToRelative.catalog_root
        absolute_path = "/catalog/root/path/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        relative_path = Transaction._abs_txn_meta_path_to_relative(
            catalog_root, absolute_path
        )
        assert (
            relative_path
            == "namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        )

    def test_abs_to_relative_same_paths(self):
        catalog_root = TestAbsToRelative.catalog_root
        absolute_path = TestAbsToRelative.catalog_root
        with pytest.raises(
            ValueError,
            match="Target and root are identical, but expected target to be a child of root.",
        ):
            Transaction._abs_txn_meta_path_to_relative(catalog_root, absolute_path)

    def test_abs_to_relative_root_with_trailing_slash(self):
        catalog_root = "/catalog/root/path/"
        absolute_path = "/catalog/root/path/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        relative_path = Transaction._abs_txn_meta_path_to_relative(
            catalog_root, absolute_path
        )
        assert (
            relative_path
            == "namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        )

    def test_abs_to_relative_bad_root(self):
        catalog_root = TestAbsToRelative.catalog_root
        absolute_path = "/cat/rt/pth/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        with pytest.raises(ValueError, match="Expected target to be a child of root."):
            Transaction._abs_txn_meta_path_to_relative(catalog_root, absolute_path)

    def test_abs_to_relative_empty_path(self):
        with pytest.raises(ValueError, match="Expected target to be a child of root."):
            Transaction._abs_txn_meta_path_to_relative("", "/lorem/ipsum")
        with pytest.raises(ValueError, match="Expected target to be a child of root."):
            Transaction._abs_txn_meta_path_to_relative("/lorem/ipsum/", "")

    # Test cases for the relativize_operation_paths function
    def test_relativize_metafile_write_paths(self):
        catalog_root = "/catalog/root"
        absolute_paths = [
            "/catalog/root/path/to/metafile1.mpk",
            "/catalog/root/path/to/metafile2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        expected_relative_paths = [
            "path/to/metafile1.mpk",
            "path/to/metafile2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=dest_metafile,
        )
        # use replace method as setter
        transaction_operation.metafile_write_paths = absolute_paths
        # Create a transaction and relativize paths
        transaction = Transaction.of(
            txn_type=TransactionType.APPEND, txn_operations=[transaction_operation]
        )
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        # Verify the paths have been correctly relativized
        assert transaction_operation.metafile_write_paths == expected_relative_paths

    def test_relativize_locator_write_paths(self):
        catalog_root = "/catalog/root"
        absolute_paths = [
            "/catalog/root/path/to/loc1.mpk",
            "/catalog/root/path/to/loc2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        expected_relative_paths = [
            "path/to/loc1.mpk",
            "path/to/loc2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=dest_metafile,
        )
        # use replace as setter
        transaction_operation.locator_write_paths = absolute_paths
        # Create a transaction and relativize paths
        transaction = Transaction.of(
            txn_type=TransactionType.APPEND, txn_operations=[transaction_operation]
        )
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        # Verify the paths have been correctly relativized
        assert transaction_operation.locator_write_paths == expected_relative_paths

    def test_relativize_metafile_and_locator_paths(self):
        catalog_root = "/meta_catalog/root_dir/a/b/c"
        meta_absolute_paths = [
            "/meta_catalog/root_dir/a/b/c/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "/meta_catalog/root_dir/a/b/c/namespace/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "/meta_catalog/root_dir/a/b/c/namespace/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        loc_absolute_paths = [
            "/meta_catalog/root_dir/a/b/c/d/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "/meta_catalog/root_dir/a/b/c/e/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "/meta_catalog/root_dir/a/b/c/f/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        meta_relative_paths = [
            "namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "namespace/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "namespace/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        loc_relative_paths = [
            "d/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "e/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "f/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=dest_metafile,
        )
        # use replace as setter
        transaction_operation.metafile_write_paths = meta_absolute_paths
        transaction_operation.locator_write_paths = loc_absolute_paths
        # Create a transaction and relativize paths
        transaction = Transaction.of(
            txn_type=TransactionType.APPEND, txn_operations=[transaction_operation]
        )
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        # Verify the paths have been correctly relativized
        assert (
            transaction_operation.metafile_write_paths == meta_relative_paths
        ), f"Expected: {meta_relative_paths}, but got: {transaction_operation.metafile_write_paths}"
        assert (
            transaction_operation.locator_write_paths == loc_relative_paths
        ), f"Expected: {loc_relative_paths}, but got: {transaction_operation.locator_write_paths}"

    def test_multiple_operations_relativize_paths(self):
        catalog_root = "/catalog/root"
        meta_absolute_paths = [
            "/catalog/root/path/to/metafile1.mpk",
            "/catalog/root/path/to/metafile2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        loc_absolute_paths = [
            "/catalog/root/path/to/loc1.mpk",
            "/catalog/root/path/to/loc2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        meta_expected_relative_paths = [
            "path/to/metafile1.mpk",
            "path/to/metafile2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        loc_expected_relative_paths = [
            "path/to/loc1.mpk",
            "path/to/loc2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operations = []
        for i in range(11):
            transaction_operation = TransactionOperation.of(
                operation_type=TransactionOperationType.CREATE,
                dest_metafile=dest_metafile,
            )
            transaction_operation.metafile_write_paths = meta_absolute_paths
            transaction_operation.locator_write_paths = loc_absolute_paths
            transaction_operations.append(transaction_operation)
        # Create a transaction and relativize paths
        transaction = Transaction.of(
            txn_type=TransactionType.APPEND, txn_operations=transaction_operations
        )
        for operation in transaction_operations:
            transaction.relativize_operation_paths(operation, catalog_root)
        # Verify the paths have been correctly relativized
        for operation in transaction_operations:
            assert operation.metafile_write_paths == meta_expected_relative_paths
            assert operation.locator_write_paths == loc_expected_relative_paths

    def test_empty_metafile_and_locator_write_paths(self):
        catalog_root = "/catalog/root"
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=Metafile({"id": "dummy_metafile_id"}),
        )
        # Empty paths
        transaction_operation.metafile_write_paths = []
        transaction_operation.locator_write_paths = []
        transaction = Transaction.of(
            txn_type=TransactionType.APPEND, txn_operations=[transaction_operation]
        )
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        assert transaction_operation.metafile_write_paths == []
        assert transaction_operation.locator_write_paths == []

    def test_large_number_of_paths(self):
        catalog_root = "/catalog/root"
        absolute_paths = [f"/catalog/root/path/to/file{i}.mpk" for i in range(5000)]
        expected_paths = [f"path/to/file{i}.mpk" for i in range(5000)]
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=Metafile({"id": "dummy_metafile_id"}),
        )
        transaction_operation.metafile_write_paths = absolute_paths
        transaction = Transaction.of(
            txn_type=TransactionType.APPEND, txn_operations=[transaction_operation]
        )
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        assert transaction_operation.metafile_write_paths == expected_paths

    def test_large_number_of_paths_multi_ops(self):
        catalog_root = "/catalog/root"
        absolute_paths = [f"/catalog/root/path/to/file{i}.mpk" for i in range(1000)]
        expected_paths = [f"path/to/file{i}.mpk" for i in range(1000)]

        # Different operation types to test
        operation_types = [
            TransactionOperationType.CREATE,
            # TransactionOperationType.UPDATE,
            TransactionOperationType.DELETE,
            TransactionOperationType.READ_EXISTS,
            TransactionOperationType.READ_LATEST,
            TransactionOperationType.READ_CHILDREN,
            TransactionOperationType.READ_SIBLINGS,
        ]

        # Different transaction types to test
        txn_types = [
            TransactionType.APPEND,
            TransactionType.ALTER,
            TransactionType.DELETE,
            TransactionType.OVERWRITE,
            TransactionType.READ,
            TransactionType.RESTATE,
        ]

        for txn_type in txn_types:
            transaction_ops = []
            for op_type in operation_types:
                transaction_operation = TransactionOperation.of(
                    operation_type=op_type,
                    dest_metafile=Metafile({"id": "dummy_metafile_id"}),
                )
                transaction_operation.metafile_write_paths = absolute_paths
                transaction_ops.append(transaction_operation)
            transaction = Transaction.of(
                txn_type=txn_type, txn_operations=[transaction_operation]
            )
            transaction.relativize_operation_paths(transaction_operation, catalog_root)
            # Assert paths are relativized correctly
            assert (
                transaction_operation.metafile_write_paths == expected_paths
            ), f"Failed for transaction type {txn_type} and operation type {op_type}"



from ray.data import from_items
from typing import Any
import pytest
import fsspec
from fsspec import AbstractFileSystem
from ray.data.datasource import FilenameProvider
from deltacat.types.media import ContentType
import ray


class TestDatasetToFile:

    BASE_PATH = "/tmp"
    SUB_PATH = "abcd"

    @pytest.fixture(autouse=True, scope="module")
    def ensure_ray_down(self):
        # ray.data fails when ray is instantiated in local mode
        ray.shutdown()

    @pytest.fixture(scope="module")
    def mock_dataset(self):
        return from_items([{"col1": i, "col2": i * 2} for i in range(1000)])

    @pytest.fixture(scope="module")
    def mock_filename_provider(self):
        class MockFilenameProvider(FilenameProvider):
            def get_filename_for_block(
                self, block: Any, task_index: int, block_index: int
            ) -> str:
                return TestDatasetToFile.SUB_PATH

        return MockFilenameProvider()

    def test_parquet_sanity(self, mock_dataset, mock_filename_provider):
        from deltacat.utils.ray_utils.dataset import dataset_to_file

        fs: AbstractFileSystem = fsspec.filesystem("local")

        dataset_to_file(
            mock_dataset,
            self.BASE_PATH,
            file_system=fs,
            block_path_provider=mock_filename_provider,
        )

        file_expected_at = f"{self.BASE_PATH}/{self.SUB_PATH}"
        assert fs.exists(file_expected_at), "file was not written"
        fs.delete(file_expected_at)

    def test_csv_sanity(self, mock_dataset, mock_filename_provider):
        from deltacat.utils.ray_utils.dataset import dataset_to_file

        fs: AbstractFileSystem = fsspec.filesystem("local")

        dataset_to_file(
            mock_dataset,
            self.BASE_PATH,
            file_system=fs,
            block_path_provider=mock_filename_provider,
            content_type=ContentType.CSV.value,
        )

        file_expected_at = f"{self.BASE_PATH}/{self.SUB_PATH}"
        assert fs.exists(file_expected_at), "file was not written"
        fs.delete(file_expected_at)



import shutil
import tempfile
import deltacat as dc


class TestDeltaCAT:
    @classmethod
    def setup_class(cls):
        cls.temp_dir_1 = tempfile.mkdtemp()
        cls.temp_dir_2 = tempfile.mkdtemp()
        # Initialize DeltaCAT with two local catalogs.
        dc.put("test_catalog_1", root=cls.temp_dir_1)
        dc.put("test_catalog_2", root=cls.temp_dir_2)

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.temp_dir_1)
        shutil.rmtree(cls.temp_dir_2)

    def test_cross_catalog_namespace_copy(self):
        # Given two empty DeltaCAT catalogs.
        # When a namespace is copied across catalogs.
        namespace_src = dc.put("test_catalog_1/test_namespace")
        namespace_dst = dc.copy(
            "test_catalog_1/test_namespace",
            "test_catalog_2",
        )
        # Expect the catalog namespace created in each catalog
        # method to be equivalent and equal to the source namespace.
        assert namespace_src.equivalent_to(namespace_dst)
        assert namespace_src == namespace_dst

        # When each catalog namespace is fetched explicitly
        # Expect them to be equivalent but not equal
        # (due to different metafile IDs).
        actual_namespace_src = dc.get("test_catalog_1/test_namespace")
        actual_namespace_dst = dc.get("test_catalog_2/test_namespace")
        assert actual_namespace_src.equivalent_to(actual_namespace_dst)
        assert not actual_namespace_src == actual_namespace_dst



from __future__ import annotations
from typing import Dict, List
from deltacat.compute.converter.model.convert_input_files import ConvertInputFiles


class ConvertInput(Dict):
    @staticmethod
    def of(
        convert_input_files,
        convert_task_index,
        iceberg_table_warehouse_prefix,
        identifier_fields,
        compact_small_files,
        enforce_primary_key_uniqueness,
        position_delete_for_multiple_data_files,
        max_parallel_data_file_download,
        s3_file_system,
    ) -> ConvertInput:

        result = ConvertInput()
        result["convert_input_files"] = convert_input_files
        result["convert_task_index"] = convert_task_index
        result["identifier_fields"] = identifier_fields
        result["iceberg_table_warehouse_prefix"] = iceberg_table_warehouse_prefix
        result["compact_small_files"] = compact_small_files
        result["enforce_primary_key_uniqueness"] = enforce_primary_key_uniqueness
        result[
            "position_delete_for_multiple_data_files"
        ] = position_delete_for_multiple_data_files
        result["max_parallel_data_file_download"] = max_parallel_data_file_download
        result["s3_file_system"] = s3_file_system

        return result

    @property
    def convert_input_files(self) -> ConvertInputFiles:
        return self["convert_input_files"]

    @property
    def identifier_fields(self) -> List[str]:
        return self["identifier_fields"]

    @property
    def convert_task_index(self) -> int:
        return self["convert_task_index"]

    @property
    def iceberg_table_warehouse_prefix(self) -> str:
        return self["iceberg_table_warehouse_prefix"]

    @property
    def compact_small_files(self) -> bool:
        return self["compact_small_files"]

    @property
    def enforce_primary_key_uniqueness(self) -> bool:
        return self["enforce_primary_key_uniqueness"]

    @property
    def position_delete_for_multiple_data_files(self) -> bool:
        return self["position_delete_for_multiple_data_files"]

    @property
    def max_parallel_data_file_download(self) -> int:
        return self["max_parallel_data_file_download"]

    @property
    def s3_file_system(self):
        return self["s3_file_system"]



import pyarrow.compute as pc

import deltacat.compute.converter.utils.iceberg_columns as sc
import pyarrow as pa

from collections import defaultdict
import ray
import logging
from deltacat.compute.converter.model.convert_input import ConvertInput
from deltacat.compute.converter.steps.dedupe import dedupe_data_files
from deltacat.compute.converter.utils.s3u import upload_table_with_retry
from deltacat.compute.converter.utils.io import (
    download_data_table_and_append_iceberg_columns,
)
from deltacat.compute.converter.utils.converter_session_utils import (
    partition_value_record_to_partition_value_string,
)

from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote
def convert(convert_input: ConvertInput):
    convert_input_files = convert_input.convert_input_files
    convert_task_index = convert_input.convert_task_index
    iceberg_table_warehouse_prefix = convert_input.iceberg_table_warehouse_prefix
    identifier_fields = convert_input.identifier_fields
    compact_small_files = convert_input.compact_small_files
    position_delete_for_multiple_data_files = (
        convert_input.position_delete_for_multiple_data_files
    )
    max_parallel_data_file_download = convert_input.max_parallel_data_file_download
    s3_file_system = convert_input.s3_file_system
    if not position_delete_for_multiple_data_files:
        raise NotImplementedError(
            f"Distributed file level position delete compute is not supported yet"
        )
    if compact_small_files:
        raise NotImplementedError(f"Compact previous position delete not supported yet")

    logger.info(f"Starting convert task index: {convert_task_index}")

    applicable_data_files = convert_input_files.applicable_data_files
    applicable_equality_delete_files = (
        convert_input_files.applicable_equality_delete_files
    )
    all_data_files_for_this_bucket = convert_input_files.all_data_files_for_dedupe

    partition_value_str = partition_value_record_to_partition_value_string(
        convert_input_files.partition_value
    )
    partition_value = convert_input_files.partition_value
    iceberg_table_warehouse_prefix_with_partition = (
        f"{iceberg_table_warehouse_prefix}/{partition_value_str}"
    )
    enforce_primary_key_uniqueness = convert_input.enforce_primary_key_uniqueness
    total_pos_delete_table = []
    if applicable_equality_delete_files:
        (
            pos_delete_after_converting_equality_delete
        ) = compute_pos_delete_with_limited_parallelism(
            data_files_list=applicable_data_files,
            identifier_columns=identifier_fields,
            equality_delete_files_list=applicable_equality_delete_files,
            iceberg_table_warehouse_prefix_with_partition=iceberg_table_warehouse_prefix_with_partition,
            max_parallel_data_file_download=max_parallel_data_file_download,
            s3_file_system=s3_file_system,
        )
        if pos_delete_after_converting_equality_delete:
            total_pos_delete_table.append(pos_delete_after_converting_equality_delete)

    if enforce_primary_key_uniqueness:
        data_files_to_dedupe = get_additional_applicable_data_files(
            all_data_files=all_data_files_for_this_bucket,
            data_files_downloaded=applicable_data_files,
        )
        pos_delete_after_dedupe = dedupe_data_files(
            data_file_to_dedupe=data_files_to_dedupe,
            identify_column_name_concatenated=identifier_fields[0],
            identifier_columns=identifier_fields,
            merge_sort_column=sc._ORDERED_RECORD_IDX_COLUMN_NAME,
        )
        total_pos_delete_table.append(pos_delete_after_dedupe)

    total_pos_delete = pa.concat_tables(total_pos_delete_table)
    to_be_added_files_list = upload_table_with_retry(
        table=total_pos_delete,
        s3_url_prefix=iceberg_table_warehouse_prefix_with_partition,
        s3_table_writer_kwargs={},
        s3_file_system=s3_file_system,
    )

    to_be_delete_files_dict = defaultdict()
    if applicable_equality_delete_files:
        to_be_delete_files_dict[partition_value] = [
            equality_delete_file[1]
            for equality_delete_file in applicable_equality_delete_files
        ]
    to_be_added_files_dict = defaultdict()
    to_be_added_files_dict[partition_value] = to_be_added_files_list
    return (to_be_delete_files_dict, to_be_added_files_dict)


def get_additional_applicable_data_files(all_data_files, data_files_downloaded):
    data_file_to_dedupe = all_data_files
    if data_files_downloaded:
        data_file_to_dedupe = list(set(all_data_files) - set(data_files_downloaded))
    return data_file_to_dedupe


def filter_rows_to_be_deleted(
    equality_delete_table, data_file_table, identifier_columns
):
    identifier_column = identifier_columns[0]
    if equality_delete_table and data_file_table:
        equality_deletes = pc.is_in(
            data_file_table[identifier_column],
            equality_delete_table[identifier_column],
        )
        position_delete_table = data_file_table.filter(equality_deletes)
        logger.info(f"positional_delete_table:{position_delete_table.to_pydict()}")
        logger.info(f"data_file_table:{data_file_table.to_pydict()}")
        logger.info(
            f"length_pos_delete_table, {len(position_delete_table)}, length_data_table:{len(data_file_table)}"
        )
    return position_delete_table


def compute_pos_delete_converting_equality_deletes(
    equality_delete_table,
    data_file_table,
    identifier_columns,
    iceberg_table_warehouse_prefix_with_partition,
    s3_file_system,
):
    new_position_delete_table = filter_rows_to_be_deleted(
        data_file_table=data_file_table,
        equality_delete_table=equality_delete_table,
        identifier_columns=identifier_columns,
    )
    if new_position_delete_table:
        logger.info(
            f"Length of position delete table after converting from equality deletes:{len(new_position_delete_table)}"
        )
    else:
        return None
    return new_position_delete_table


def download_bucketed_table(data_files, equality_delete_files):
    from deltacat.utils.pyarrow import s3_file_to_table

    compacted_table = s3_file_to_table(
        [data_file.file_path for data_file in data_files]
    )
    equality_delete_table = s3_file_to_table(
        [eq_file.file_path for eq_file in equality_delete_files]
    )
    return compacted_table, equality_delete_table


def compute_pos_delete_with_limited_parallelism(
    data_files_list,
    identifier_columns,
    equality_delete_files_list,
    iceberg_table_warehouse_prefix_with_partition,
    max_parallel_data_file_download,
    s3_file_system,
):
    for data_files, equality_delete_files in zip(
        data_files_list, equality_delete_files_list
    ):
        data_table_total = []
        for data_file in data_files:
            data_table = download_data_table_and_append_iceberg_columns(
                data_files=data_file[1],
                columns_to_download=identifier_columns,
                additional_columns_to_append=[
                    sc._FILE_PATH_COLUMN_NAME,
                    sc._ORDERED_RECORD_IDX_COLUMN_NAME,
                ],
                sequence_number=data_file[0],
            )
            data_table_total.append(data_table)
        data_table_total = pa.concat_tables(data_table_total)

        equality_delete_table_total = []
        for equality_delete in equality_delete_files:
            equality_delete_table = download_data_table_and_append_iceberg_columns(
                data_files=equality_delete[1],
                columns_to_download=identifier_columns,
            )
            equality_delete_table_total.append(equality_delete_table)
        equality_delete_table_total = pa.concat_tables(equality_delete_table_total)

    new_pos_delete_table = compute_pos_delete_converting_equality_deletes(
        equality_delete_table=equality_delete_table_total,
        data_file_table=data_table_total,
        iceberg_table_warehouse_prefix_with_partition=iceberg_table_warehouse_prefix_with_partition,
        identifier_columns=identifier_columns,
        s3_file_system=s3_file_system,
    )
    if not new_pos_delete_table:
        logger.info("No records deleted based on equality delete converstion")

    logger.info(
        f"Number of records to delete based on equality delete convertion:{len(new_pos_delete_table)}"
    )
    return new_pos_delete_table



from __future__ import annotations

import os
import copy
import time
import uuid
import posixpath
from pathlib import PosixPath
import threading
from collections import defaultdict

from itertools import chain
from typing import Optional, List, Union, Tuple

import msgpack
import pyarrow.fs

from deltacat.constants import (
    TXN_DIR_NAME,
    TXN_PART_SEPARATOR,
    RUNNING_TXN_DIR_NAME,
    FAILED_TXN_DIR_NAME,
    SUCCESS_TXN_DIR_NAME,
    NANOS_PER_SEC,
)
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.types import (
    TransactionOperationType,
    TransactionType,
)
from deltacat.storage.model.metafile import (
    Metafile,
    MetafileRevisionInfo,
)
from deltacat.utils.filesystem import (
    resolve_path_and_filesystem,
    list_directory,
)


class TransactionTimeProvider:
    """
    Provider interface for transaction start and end times. An ideal
    transaction time provider is externally consistent (e.g.,
    https://cloud.google.com/spanner/docs/true-time-external-consistency),
    such that:
      1. A transaction start time is never less than a previously completed
      transaction's end time.
      2. A transaction end time is never less than an in-progress
      transaction's start time.
      3. Every transaction has a unique start and end time.
      4. Start/end time assignment is non-blocking.
    """

    def start_time(self) -> int:
        raise NotImplementedError("start_time not implemented")

    def end_time(self) -> int:
        raise NotImplementedError("end_time not implemented")


class TransactionSystemTimeProvider(TransactionTimeProvider):
    """
    A local transaction time provider that returns the current system clock
    epoch time in nanoseconds. Ensures that all local transaction start
    times are greater than all last known end times, and that all known end
    times are no less than all last known start time across all local threads
    using this time provider.

    Note that this time provider gives no external consistency guarantees due
    to potential clock skew between distributed nodes writing to the same
    catalog, and is only recommended for use with local catalogs.
    """

    last_known_start_times = defaultdict(int)
    last_known_end_times = defaultdict(int)

    # don't wait more than 60 seconds for the system clock to catch up
    # between transactions (assumed to be indicative of a larger system
    # clock change made between transactions)
    max_sync_wait_time = 60 * NANOS_PER_SEC

    def start_time(self) -> int:
        """
        Gets the current system time in nanoseconds since the epoch. Ensures
        that the start time returned is greater than the last known end time
        recorded at the time this method is invoked.
        :return: Current epoch time in nanoseconds.
        """
        # ensure serial transactions in a single process have start times after
        # the last known end time
        last_known_end_times = self.last_known_end_times.values() or [0]
        max_known_end_time = max(last_known_end_times)

        elapsed_start_time = time.monotonic_ns()
        current_time = time.time_ns()
        while current_time <= max_known_end_time:
            elapsed_time = time.monotonic_ns() - elapsed_start_time
            if elapsed_time > self.max_sync_wait_time:
                raise TimeoutError(
                    f"Failed to sync cross-transaction system clock time after "
                    f"{self.max_sync_wait_time / NANOS_PER_SEC} seconds, "
                    f"aborting."
                )
            time.sleep(0.000001)
            current_time = time.time_ns()

        # update the current thread's last known end time
        pid = os.getpid()
        tid = threading.current_thread().ident
        current_thread_time_key = (pid, tid)
        self.last_known_end_times[current_thread_time_key] = current_time

        return current_time

    def end_time(self) -> int:
        """
        Gets the current system time in nanoseconds since the epoch. Ensures
        that the end time returned is no less than the last known start time
        recorded at the time this method is invoked.
        :return: Current epoch time in nanoseconds.
        """
        # ensure serial transactions in a single process have end times no less
        # than the last known start time
        last_known_start_times = self.last_known_start_times.values() or [0]
        last_start_time = max(last_known_start_times)

        elapsed_start_time = time.monotonic_ns()
        current_time = time.time_ns()
        while current_time < last_start_time:
            elapsed_time = time.monotonic_ns() - elapsed_start_time
            if elapsed_time > self.max_sync_wait_time:
                raise TimeoutError(
                    f"Failed to sync cross-transaction system clock time after "
                    f"{self.max_sync_wait_time / NANOS_PER_SEC} seconds, "
                    f"aborting."
                )
            time.sleep(0.000001)
            current_time = time.time_ns()

        # update the current thread's last known end time
        pid = os.getpid()
        tid = threading.current_thread().ident
        current_thread_time_key = (pid, tid)
        self.last_known_start_times[current_thread_time_key] = current_time

        return current_time


class TransactionOperation(dict):
    """
    Base class for DeltaCAT transaction operations against individual metafiles.
    """

    @staticmethod
    def of(
        operation_type: Optional[TransactionOperationType],
        dest_metafile: Metafile,
        src_metafile: Optional[Metafile] = None,
        read_limit: Optional[int] = None,
    ) -> TransactionOperation:
        if not dest_metafile:
            raise ValueError("Transaction operations must have a destination metafile.")
        if operation_type == TransactionOperationType.UPDATE:
            if not src_metafile:
                raise ValueError(
                    "UPDATE transaction operations must have a source metafile."
                )
            elif type(dest_metafile) is not type(src_metafile):
                raise ValueError(
                    f"Source metafile type `{type(src_metafile)}` is not "
                    f"equal to dest metafile type `{type(dest_metafile)}`."
                )
        elif src_metafile:
            raise ValueError(
                "Only UPDATE transaction operations may have a source metafile."
            )
        if operation_type.is_write_operation() and read_limit:
            raise ValueError("Only READ transaction operations may have a read limit.")
        txn_op = TransactionOperation()
        txn_op.type = operation_type
        txn_op.dest_metafile = dest_metafile
        txn_op.src_metafile = src_metafile
        txn_op.read_limit = read_limit
        return txn_op

    @property
    def type(self) -> TransactionOperationType:
        """
        Returns the type of the transaction operation.
        """
        return TransactionOperationType(self["type"])

    @type.setter
    def type(self, txn_op_type: TransactionOperationType):
        self["type"] = txn_op_type

    @property
    def dest_metafile(self) -> Metafile:
        """
        Returns the metafile that is the target of this transaction operation.
        """
        return self["dest_metafile"]

    @dest_metafile.setter
    def dest_metafile(self, metafile: Metafile):
        self["dest_metafile"] = metafile

    @property
    def src_metafile(self) -> Optional[Metafile]:
        """
        Returns the metafile that is the source of this transaction operation.
        """
        return self["src_metafile"]

    @src_metafile.setter
    def src_metafile(self, src_metafile: Optional[Metafile]):
        self["src_metafile"] = src_metafile

    @property
    def read_limit(self) -> Optional[int]:
        """
        Returns the read limit for this transaction operation.
        """
        return self.get("read_limit")

    @read_limit.setter
    def read_limit(self, read_limit: Optional[int]):
        self["read_limit"] = read_limit

    @property
    def metafile_write_paths(self) -> List[str]:
        return self.get("metafile_write_paths") or []

    @property
    def locator_write_paths(self) -> List[str]:
        return self.get("locator_write_paths") or []

    def append_metafile_write_path(self, write_path: str):
        metafile_write_paths = self.get("metafile_write_paths")
        if not metafile_write_paths:
            metafile_write_paths = self["metafile_write_paths"] = []
        metafile_write_paths.append(write_path)

    def append_locator_write_path(self, write_path: str):
        locator_write_paths = self.get("locator_write_paths")
        if not locator_write_paths:
            locator_write_paths = self["locator_write_paths"] = []
        locator_write_paths.append(write_path)

    @metafile_write_paths.setter
    def metafile_write_paths(self, write_paths: List[str]) -> None:
        self["metafile_write_paths"] = write_paths

    @locator_write_paths.setter
    def locator_write_paths(self, write_paths: List[str]):
        self["locator_write_paths"] = write_paths


class TransactionOperationList(List[TransactionOperation]):
    @staticmethod
    def of(items: List[TransactionOperation]) -> TransactionOperationList:
        typed_items = TransactionOperationList()
        for item in items:
            if item is not None and not isinstance(item, TransactionOperation):
                item = TransactionOperation(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, TransactionOperation):
            self[item] = val = TransactionOperation(val)
        return val


class Transaction(dict):
    """
    Base class for DeltaCAT transactions.
    """

    @staticmethod
    def of(
        txn_type: TransactionType,
        txn_operations: Optional[TransactionOperationList],
    ) -> Transaction:
        operation_types = set([op.type for op in txn_operations])
        if txn_type == TransactionType.READ:
            if operation_types - TransactionOperationType.read_operations():
                raise ValueError(
                    "Only READ transaction operation types may be specified as "
                    "part of a READ transaction."
                )
        elif (
            len(operation_types) == 1
            and TransactionOperationType.CREATE in operation_types
        ):
            if txn_type != TransactionType.APPEND:
                raise ValueError(
                    "Transactions with only CREATE operations must be "
                    "specified as part of an APPEND transaction."
                )
        elif TransactionOperationType.DELETE in operation_types:
            if txn_type != TransactionType.DELETE:
                raise ValueError(
                    "DELETE transaction operations must be specified as part "
                    "of a DELETE transaction."
                )
        elif TransactionOperationType.UPDATE in operation_types and txn_type not in {
            TransactionType.ALTER,
            TransactionType.RESTATE,
            TransactionType.OVERWRITE,
        }:
            raise ValueError(
                "Transactions with UPDATE operations must be specified "
                "as part of an ALTER, RESTATE, or OVERWRITE transaction."
            )
        transaction = Transaction()
        transaction.type = txn_type
        transaction.operations = txn_operations
        return transaction

    @staticmethod
    def read_end_time(
        path: str,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
    ) -> Optional[int]:
        """
        Returns the end time of the transaction, or None if the transaction
        log file does not exist.
        :param path: Transaction log path to read.
        :param filesystem: File system to use for reading the Transaction file.
        :return: Deserialized object from the Transaction file.
        """
        # TODO(pdames): Validate that input file path is a valid txn log.
        if not filesystem:
            path, filesystem = resolve_path_and_filesystem(path, filesystem)
        file_info_and_sizes = list_directory(
            path=path,
            filesystem=filesystem,
            ignore_missing_path=True,
        )
        end_time = None
        if file_info_and_sizes:
            if len(file_info_and_sizes) > 1:
                raise ValueError(
                    f"Expected to find only one transaction log at {path}, "
                    f"but found {len(file_info_and_sizes)}"
                )
            end_time = Transaction._parse_end_time(file_info_and_sizes[0][0])
        return end_time

    @staticmethod
    def _parse_end_time(txn_log_file_name_or_path: str) -> int:
        return int(posixpath.basename(txn_log_file_name_or_path))

    @classmethod
    def read(
        cls,
        path: str,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
    ) -> Transaction:
        """
        Read a Transaction file and return the deserialized object.
        :param path: Transaction file path to read.
        :param filesystem: File system to use for reading the Transaction file.
        :return: Deserialized object from the Transaction file.
        """
        if not filesystem:
            path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_input_stream(path) as file:
            binary = file.readall()
        obj = cls(**msgpack.loads(binary))
        return obj

    @property
    def id(self) -> Optional[str]:
        """
        Returns this transaction's unique ID assigned at commit start time, or
        None if the unique ID has not yet been assigned.
        """
        _id = self.get("id")
        if not _id and self.start_time:
            _id = self["id"] = f"{self.start_time}{TXN_PART_SEPARATOR}{uuid.uuid4()}"
        return _id

    @property
    def type(self) -> TransactionType:
        """
        Returns the type of the transaction.
        """
        return TransactionType(self["type"])

    @type.setter
    def type(self, txn_type: TransactionType):
        self["type"] = txn_type

    @property
    def operations(self) -> TransactionOperationList:
        """
        Returns the list of transaction operations.
        """
        return TransactionOperationList(self["operations"])

    @operations.setter
    def operations(self, operations: TransactionOperationList):
        self["operations"] = operations

    @property
    def start_time(self) -> Optional[int]:
        """
        Returns the start time of the transaction.
        """
        return self.get("start_time")

    @property
    def end_time(self) -> Optional[int]:
        """
        Returns the end time of the transaction.
        """
        return self.get("end_time")

    def _mark_start_time(self, time_provider: TransactionTimeProvider) -> int:
        """
        Sets the start time of the transaction using the given
        TransactionTimeProvider. Raises a runtime error if the transaction
        start time has already been set by a previous commit.
        """
        if self.get("start_time"):
            raise RuntimeError("Cannot restart a previously started transaction.")
        start_time = self["start_time"] = time_provider.start_time()
        return start_time

    def _mark_end_time(self, time_provider: TransactionTimeProvider) -> int:
        """
        Sets the end time of the transaction using the given
        TransactionTimeProvider. Raises a runtime error if the transaction end
        time has already been set by a previous commit, or if the transaction
        start time has not been set.
        """
        if not self.get("start_time"):
            raise RuntimeError("Cannot end an unstarted transaction.")
        if self.get("end_time"):
            raise RuntimeError("Cannot end a completed transaction.")
        end_time = self["end_time"] = time_provider.end_time()
        return end_time

    @staticmethod
    def _abs_txn_meta_path_to_relative(root: str, target: str) -> str:
        """
        Takes an absolute root directory path and target absolute path to
        relativize with respect to the root directory. Returns the target
        path relative to the root directory path. Raises an error if the
        target path is not contained in the given root directory path, if
        either path is not an absolute path, or if the target path is equal
        to the root directory path.
        """
        root_path = PosixPath(root)
        target_path = PosixPath(target)
        # TODO (martinezdavid): Check why is_absolute() fails for certain Delta paths
        # if not root_path.is_absolute() or not target_path.is_absolute():
        #     raise ValueError("Both root and target must be absolute paths.")
        if root_path == target_path:
            raise ValueError(
                "Target and root are identical, but expected target to be a child of root."
            )
        try:
            relative_path = target_path.relative_to(root_path)
        except ValueError:
            raise ValueError("Expected target to be a child of root.")
        return str(relative_path)

    def relativize_operation_paths(
        self, operation: TransactionOperation, catalog_root: str
    ) -> None:
        """
        Converts all absolute paths in an operation to relative paths
        with respect to the catalog root directory.
        """
        # handle metafile paths
        if operation.metafile_write_paths:
            metafile_write_paths = [
                Transaction._abs_txn_meta_path_to_relative(catalog_root, path)
                for path in operation.metafile_write_paths
            ]
            operation.metafile_write_paths = metafile_write_paths
        # handle locator paths
        if operation.locator_write_paths:
            locator_write_paths = [
                Transaction._abs_txn_meta_path_to_relative(catalog_root, path)
                for path in operation.locator_write_paths
            ]
            operation.locator_write_paths = locator_write_paths

    def to_serializable(self, catalog_root) -> Transaction:
        """
        Prepare the object for serialization by converting any non-serializable
        types to serializable types. May also run any required pre-write
        validations on the serialized or deserialized object.
        :return: a serializable version of the object
        """
        serializable = copy.deepcopy(self)
        # remove all src/dest metafile contents except IDs and locators to
        # reduce file size (they can be reconstructed from their corresponding
        # files as required).
        for operation in serializable.operations:
            # Sanity check that IDs exist on source and dest metafiles
            if operation.dest_metafile and operation.dest_metafile.id is None:
                raise ValueError(
                    f"Transaction operation ${operation} dest metafile does "
                    f"not have ID: ${operation.dest_metafile}"
                )
            if operation.src_metafile and operation.src_metafile.id is None:
                raise ValueError(
                    f"Transaction operation ${operation} src metafile does "
                    f"not have ID: ${operation.src_metafile}"
                )
            # relativize after checking that dest and src metafiles are valid
            self.relativize_operation_paths(operation, catalog_root)
            operation.dest_metafile = {
                "id": operation.dest_metafile.id,
                "locator": operation.dest_metafile.locator,
                "locator_alias": operation.dest_metafile.locator_alias,
            }
            if operation.src_metafile:
                operation.src_metafile = {
                    "id": operation.src_metafile.id,
                    "locator": operation.src_metafile.locator,
                    "locator_alias": operation.src_metafile.locator_alias,
                }
        # TODO(pdames): Ensure that all file paths recorded are relative to the
        #  catalog root.
        return serializable

    @staticmethod
    def _validate_txn_log_file(success_txn_log_file: str) -> None:
        txn_log_dir_name = posixpath.basename(posixpath.dirname(success_txn_log_file))
        txn_log_parts = txn_log_dir_name.split(TXN_PART_SEPARATOR)
        # ensure that the transaction start time is valid
        try:
            start_time = int(txn_log_parts[0])
        except ValueError as e:
            raise ValueError(
                f"Transaction log file `{success_txn_log_file}` does not "
                f"contain a valid start time."
            ) from e
        # ensure that the txn uuid is valid
        txn_uuid_str = txn_log_parts[1]
        try:
            uuid.UUID(txn_uuid_str)
        except ValueError as e:
            raise OSError(
                f"Transaction log file `{success_txn_log_file}` does not "
                f"contain a valid UUID string."
            ) from e
        # ensure that the transaction end time is valid
        try:
            end_time = Transaction._parse_end_time(success_txn_log_file)
        except ValueError as e:
            raise ValueError(
                f"Transaction log file `{success_txn_log_file}` does not "
                f"contain a valid end time."
            ) from e
        # ensure transaction end time was not recorded before start time
        if end_time < start_time:
            raise OSError(
                f"Transaction end time {end_time} is earlier than start "
                f"time {start_time}! To preserve catalog integrity, the "
                f"corresponding completed transaction log at "
                f"`{success_txn_log_file}` has been removed."
            )

    def commit(
        self,
        catalog_root_dir: str,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
    ) -> Union[List[ListResult[Metafile]], Tuple[List[str], str]]:
        # TODO(pdames): allow transactions to be durably staged and resumed
        #  across multiple sessions prior to commit

        # create a new internal copy of this transaction to guard against
        # external modification and dirty state across retries
        txn = copy.deepcopy(self)

        # create the transaction directory first to telegraph that at least 1
        # transaction at this root has been attempted
        catalog_root_normalized, filesystem = resolve_path_and_filesystem(
            catalog_root_dir,
            filesystem,
        )
        txn_log_dir = posixpath.join(catalog_root_normalized, TXN_DIR_NAME)
        running_txn_log_dir = posixpath.join(txn_log_dir, RUNNING_TXN_DIR_NAME)
        filesystem.create_dir(running_txn_log_dir, recursive=True)
        failed_txn_log_dir = posixpath.join(txn_log_dir, FAILED_TXN_DIR_NAME)
        filesystem.create_dir(failed_txn_log_dir, recursive=False)
        success_txn_log_dir = posixpath.join(txn_log_dir, SUCCESS_TXN_DIR_NAME)
        filesystem.create_dir(success_txn_log_dir, recursive=False)

        # TODO(pdames): Support injection of other time providers, but ensure
        #  that ALL transactions in a catalog use the same time provider.
        time_provider = TransactionSystemTimeProvider()

        # record the transaction start time
        txn._mark_start_time(time_provider)

        if txn.type == TransactionType.READ:
            list_results = []
            for operation in self.operations:
                list_result = operation.dest_metafile.read_txn(
                    catalog_root_dir=catalog_root_normalized,
                    success_txn_log_dir=success_txn_log_dir,
                    current_txn_op=operation,
                    current_txn_start_time=txn.start_time,
                    current_txn_id=txn.id,
                    filesystem=filesystem,
                )
                list_results.append(list_result)
            return list_results
        else:
            return txn._commit_write(
                catalog_root_normalized=catalog_root_normalized,
                running_txn_log_dir=running_txn_log_dir,
                failed_txn_log_dir=failed_txn_log_dir,
                success_txn_log_dir=success_txn_log_dir,
                filesystem=filesystem,
                time_provider=time_provider,
            )

    def _commit_write(
        self,
        catalog_root_normalized: str,
        running_txn_log_dir: str,
        failed_txn_log_dir: str,
        success_txn_log_dir: str,
        filesystem: pyarrow.fs.FileSystem,
        time_provider: TransactionTimeProvider,
    ) -> Tuple[List[str], str]:
        # write the in-progress transaction log file
        running_txn_log_file_path = posixpath.join(
            running_txn_log_dir,
            self.id,
        )
        with filesystem.open_output_stream(running_txn_log_file_path) as file:
            packed = msgpack.dumps(self.to_serializable(catalog_root_normalized))
            file.write(packed)

        # write each metafile associated with the transaction
        metafile_write_paths = []
        locator_write_paths = []
        try:
            for operation in self.operations:
                operation.dest_metafile.write_txn(
                    catalog_root_dir=catalog_root_normalized,
                    success_txn_log_dir=success_txn_log_dir,
                    current_txn_op=operation,
                    current_txn_start_time=self.start_time,
                    current_txn_id=self.id,
                    filesystem=filesystem,
                )
                metafile_write_paths.extend(operation.metafile_write_paths)
                locator_write_paths.extend(operation.locator_write_paths)
                # check for conflicts with concurrent transactions
                for path in metafile_write_paths + locator_write_paths:
                    MetafileRevisionInfo.check_for_concurrent_txn_conflict(
                        success_txn_log_dir=success_txn_log_dir,
                        current_txn_revision_file_path=path,
                        filesystem=filesystem,
                    )
        except Exception:
            # write a failed transaction log file entry
            failed_txn_log_file_path = posixpath.join(
                failed_txn_log_dir,
                self.id,
            )
            with filesystem.open_output_stream(failed_txn_log_file_path) as file:
                packed = msgpack.dumps(self.to_serializable(catalog_root_normalized))
                file.write(packed)

            ###################################################################
            ###################################################################
            # failure past here telegraphs a failed transaction cleanup attempt
            ###################################################################
            ###################################################################

            # delete all files written during the failed transaction
            known_write_paths = chain.from_iterable(
                [
                    operation.metafile_write_paths + operation.locator_write_paths
                    for operation in self.operations
                ]
            )
            # TODO(pdames): Add separate janitor job to cleanup files that we
            #  either failed to add to the known write paths, or fail to delete.
            for write_path in known_write_paths:
                filesystem.delete_file(write_path)

            # delete the in-progress transaction log file entry
            filesystem.delete_file(running_txn_log_file_path)
            # failed transaction cleanup is now complete
            raise

        # record the completed transaction
        success_txn_log_file_dir = posixpath.join(
            success_txn_log_dir,
            self.id,
        )
        filesystem.create_dir(
            success_txn_log_file_dir,
            recursive=False,
        )
        end_time = self._mark_end_time(time_provider)
        success_txn_log_file_path = posixpath.join(
            success_txn_log_file_dir,
            str(end_time),
        )
        with filesystem.open_output_stream(success_txn_log_file_path) as file:
            packed = msgpack.dumps(self.to_serializable(catalog_root_normalized))
            file.write(packed)
        try:
            Transaction._validate_txn_log_file(
                success_txn_log_file=success_txn_log_file_path
            )
        except Exception as e1:
            try:
                # move the txn log from success dir to failed dir
                failed_txn_log_file_path = posixpath.join(
                    failed_txn_log_dir,
                    self.id,
                )
                filesystem.move(
                    src=success_txn_log_file_path,
                    dest=failed_txn_log_file_path,
                )
                # keep parent success txn log dir to telegraph failed validation

                ###############################################################
                ###############################################################
                # failure past here telegraphs a failed transaction validation
                # cleanup attempt
                ###############################################################
                ###############################################################
            except Exception as e2:
                raise OSError(
                    f"Failed to cleanup bad transaction log file at "
                    f"`{success_txn_log_file_path}`"
                ) from e2
            finally:
                raise RuntimeError(
                    f"Transaction validation failed. To preserve "
                    f"catalog integrity, the corresponding completed "
                    f"transaction log at `{success_txn_log_file_path}` has "
                    f"been removed."
                ) from e1
        finally:
            # delete the in-progress transaction log file entry
            filesystem.delete_file(running_txn_log_file_path)
        return metafile_write_paths, success_txn_log_file_path



import pytest
import pyarrow as pa

from deltacat.storage.model.schema import (
    Schema,
    Field,
    BASE_SCHEMA_NAME,
)


@pytest.fixture
def schema_a():
    return Schema.of(
        [
            Field.of(
                field=pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )


@pytest.fixture
def schema_b():
    return Schema.of(
        [
            Field.of(
                field=pa.field("col2", pa.string(), nullable=True),
                field_id=2,
                is_merge_key=False,
            )
        ]
    )


@pytest.fixture
def schema_c():
    return Schema.of(
        [
            Field.of(
                field=pa.field("col3", pa.float64(), nullable=False),
                field_id=3,
                is_merge_key=False,
            )
        ]
    )


@pytest.fixture
def schema_d():
    schema = Schema.of(
        [
            Field.of(
                field=pa.field("col_named", pa.int64(), nullable=True),
                field_id=4,
                is_merge_key=False,
            )
        ]
    )
    return schema


def test_of_with_dict(schema_a, schema_b):
    input_dict = {"schema_a": schema_a.arrow, "schema_b": schema_b.arrow}
    schema = Schema.of(input_dict)

    assert isinstance(schema, Schema)
    assert list(schema.subschemas.keys()) == ["schema_a", "schema_b"]
    assert list(schema.subschemas_to_field_ids.keys()) == ["schema_a", "schema_b"]
    assert schema.subschemas["schema_a"].equivalent_to(schema_a, True)
    assert schema.subschemas["schema_b"].equivalent_to(schema_b, True)


def test_of_invalid_input():
    with pytest.raises(ValueError):
        Schema.of(42)
    with pytest.raises(ValueError):
        Schema.of(["foo"])
    with pytest.raises(ValueError):
        Schema.of({"foo": [42]})


def test_insert_explicit_name(schema_a, schema_b):
    schema = Schema.of(schema_a.arrow)
    assert not schema.subschemas
    new_schema = schema.add_subschema("explicit", schema_b.arrow)
    assert not schema.subschemas
    assert "explicit" in new_schema.subschemas
    assert "explicit" in new_schema.subschemas_to_field_ids
    assert BASE_SCHEMA_NAME in new_schema.subschemas
    assert BASE_SCHEMA_NAME in new_schema.subschemas_to_field_ids
    assert new_schema.subschemas[BASE_SCHEMA_NAME].equivalent_to(schema_a)
    assert new_schema.subschemas["explicit"].equivalent_to(schema_b)


def test_insert_reserved_name_fails(schema_a, schema_b):
    schema = Schema.of(schema_a.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema(BASE_SCHEMA_NAME, schema_b.arrow)


def test_insert_duplicate_schema_name_fails(schema_a, schema_b):
    schema = Schema.of({"dupe": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.add_subschema("dupe", schema_b.arrow)


def test_insert_duplicate_field_name_fails(schema_a):
    schema = Schema.of(schema_a.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema("dupe_field_name", schema_a.arrow)


def test_insert_autofill_field_id():
    schema1 = [
        Field.of(
            field=pa.field("col1", pa.int32(), nullable=False),
            is_merge_key=True,
        )
    ]
    schema2 = [
        Field.of(
            field=pa.field("col2", pa.int32(), nullable=False),
            is_merge_key=True,
        )
    ]
    schema = Schema.of(schema1)
    new_schema = schema.add_subschema("explicit", schema2)
    assert not schema.subschemas
    assert "explicit" in new_schema.subschemas
    assert "explicit" in new_schema.subschemas_to_field_ids
    assert BASE_SCHEMA_NAME in new_schema.subschemas
    assert BASE_SCHEMA_NAME in new_schema.subschemas_to_field_ids
    assert len(new_schema.subschemas[BASE_SCHEMA_NAME].fields) == 1
    assert len(new_schema.subschemas["explicit"].fields) == 1
    assert new_schema.subschemas[BASE_SCHEMA_NAME].fields[0].id == 0
    assert new_schema.subschemas["explicit"].fields[0].id == 1
    assert new_schema.subschemas[BASE_SCHEMA_NAME].equivalent_to(Schema.of(schema1))
    schema2_with_expected_field_id = [
        Field.of(
            field=pa.field("col2", pa.int32(), nullable=False),
            field_id=1,
            is_merge_key=True,
        )
    ]
    assert new_schema.subschemas["explicit"].equivalent_to(
        Schema.of(schema2_with_expected_field_id)
    )


def test_insert_duplicate_field_name_case_insensitive_fails():
    schema1 = Schema.of(
        [
            Field.of(
                field=pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema2 = Schema.of(
        [
            Field.of(
                field=pa.field("COL1", pa.int32(), nullable=False),
                field_id=2,
                is_merge_key=True,
            )
        ]
    )
    schema = Schema.of(schema1.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema("dupe_field_name_case_insensitive", schema2.arrow)


def test_insert_duplicate_field_id_fails():
    schema1 = Schema.of(
        [
            Field.of(
                field=pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema2 = Schema.of(
        [
            Field.of(
                field=pa.field("col2", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema = Schema.of(schema1.arrow)
    with pytest.raises(ValueError):
        schema.add_subschema("dupe_field_id", schema2.arrow)


def test_update_success(schema_a, schema_b):
    schema = Schema.of({"key": schema_a.arrow})
    new_schema = schema.replace_subschema("key", schema_b.arrow)
    assert schema.subschemas["key"].equivalent_to(schema_a)
    assert new_schema.subschemas["key"].equivalent_to(schema_b)


def test_update_not_exist(schema_a):
    schema = Schema.of({"key": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.replace_subschema("nonexistent", schema_a.arrow)


def test_delete_schema_success(schema_a, schema_b):
    schema = Schema.of({"key1": schema_a.arrow, "key2": schema_b.arrow})
    new_schema = schema.delete_subschema("key1")
    assert "key1" in schema.subschemas
    assert "key2" in schema.subschemas
    assert "key1" not in new_schema.subschemas
    assert "key2" in new_schema.subschemas
    assert "key1" in schema.subschemas_to_field_ids
    assert "key2" in schema.subschemas_to_field_ids
    assert "key1" not in new_schema.subschemas_to_field_ids
    assert "key2" in new_schema.subschemas_to_field_ids


def test_delete_only_schema_fails(schema_a):
    schema = Schema.of({"key": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.delete_subschema("key")


def test_delete_schema_not_exist(schema_a):
    schema = Schema.of({"key": schema_a.arrow})
    with pytest.raises(ValueError):
        schema.delete_subschema("nonexistent")


def test_get_schemas_order(schema_a, schema_b, schema_c):
    schema = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow, "c": schema_c.arrow})
    assert list(schema.subschemas.keys()) == ["a", "b", "c"]
    assert list(schema.subschemas.values()) == [schema_a, schema_b, schema_c]
    schema = Schema.of({"a": schema_a.arrow})
    schema = schema.add_subschema("b", schema_b.arrow)
    assert list(schema.subschemas.keys()) == ["a", "b"]
    assert list(schema.subschemas.values()) == [schema_a, schema_b]
    schema = schema.add_subschema("c", schema_c.arrow)
    assert list(schema.subschemas.keys()) == ["a", "b", "c"]
    assert list(schema.subschemas.values()) == [schema_a, schema_b, schema_c]


def test_equivalent_to_same(schema_a, schema_b):
    schema1 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    schema2 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    assert schema1.equivalent_to(schema2)
    assert schema2.equivalent_to(schema1)


def test_equivalent_to_different_subschema_names(schema_a, schema_b):
    schema1 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    schema2 = Schema.of({"x": schema_a.arrow, "b": schema_b.arrow})
    assert schema1.equivalent_to(schema2)
    assert not schema1.equivalent_to(schema2, True)


def test_not_equivalent_to_different_subschema_order(schema_a, schema_b):
    schema1 = Schema.of({"a": schema_a.arrow, "b": schema_b.arrow})
    schema2 = Schema.of({"b": schema_b.arrow, "a": schema_a.arrow})
    assert not schema1.equivalent_to(schema2)


def test_equivalent_to_non_schema(schema_a):
    schema = Schema.of({"a": schema_a.arrow})
    assert not schema.equivalent_to("not a schema")


def test_equivalent_schemas_different_instances():
    """
    Edge case: ensure equivalent schemas with different instances
    are considered equivalent.
    """
    schema1 = Schema.of(
        [
            Field.of(
                pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema2 = Schema.of(
        [
            Field.of(
                pa.field("col1", pa.int32(), nullable=False),
                field_id=1,
                is_merge_key=True,
            )
        ]
    )
    schema1 = Schema.of({"key": schema1.arrow})
    schema2 = Schema.of({"key": schema2.arrow})
    assert schema1.equivalent_to(schema2)


def test_empty_schema_fails():
    with pytest.raises(ValueError):
        Schema.of({})
    with pytest.raises(ValueError):
        Schema.of([])



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import logging
import copy

import msgpack
from typing import Optional, Any, Dict, Union, List, Callable, Tuple

import pyarrow as pa
from pyarrow import ArrowInvalid

from deltacat.constants import BYTES_PER_KIBIBYTE
from deltacat.storage.model.types import (
    SchemaConsistencyType,
    SortOrder,
    NullOrder,
)
from deltacat import logs

# PyArrow Field Metadata Key used to set the Field ID when writing to Parquet.
# See: https://arrow.apache.org/docs/cpp/parquet.html#parquet-field-id
PARQUET_FIELD_ID_KEY_NAME = b"PARQUET:field_id"

# PyArrow Field Metadata Key used to store field documentation.
FIELD_DOC_KEY_NAME = b"DELTACAT:doc"

# PyArrow Field Metadata Key used to identify the field as a merge key.
FIELD_MERGE_KEY_NAME = b"DELTACAT:merge_key"

# PyArrow Field Metadata Key used to identify the field as a merge order key.
FIELD_MERGE_ORDER_KEY_NAME = b"DELTACAT:merge_order"

# PyArrow Field Metadata Key used to identify the field as an event time.
FIELD_EVENT_TIME_KEY_NAME = b"DELTACAT:event_time"

# PyArrow Field Metadata Key used to store field past default values.
FIELD_PAST_DEFAULT_KEY_NAME = b"DELTACAT:past_default"

# PyArrow Field Metadata Key used to store field future default values.
FIELD_FUTURE_DEFAULT_KEY_NAME = b"DELTACAT:future_default"

# PyArrow Field Metadata Key used to store field schema consistency type.
FIELD_CONSISTENCY_TYPE_KEY_NAME = b"DELTACAT:consistency_type"

# PyArrow Schema Metadata Key used to store schema ID value.
SCHEMA_ID_KEY_NAME = b"DELTACAT:schema_id"

# PyArrow Schema Metadata Key used to store named subschemas
SUBSCHEMAS_KEY_NAME = b"DELTACAT:subschemas"

# Set max field ID to INT32.MAX_VALUE - 200 for backwards-compatibility with
# Apache Iceberg, which sets aside this range for reserved fields
MAX_FIELD_ID_EXCLUSIVE = 2147483447

# Default name assigned to the base, unnamed single schema when a new named
# subschema is first added.
BASE_SCHEMA_NAME = "_base"

SchemaId = int
SchemaName = str
FieldId = int
FieldName = str
NestedFieldName = List[str]
FieldLocator = Union[FieldName, NestedFieldName, FieldId]

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class MergeOrder(tuple):
    @staticmethod
    def of(
        sort_order: SortOrder = SortOrder.ASCENDING,
        null_order: NullOrder = NullOrder.AT_END,
    ) -> MergeOrder:
        return MergeOrder(
            (
                sort_order,
                null_order,
            )
        )

    @property
    def sort_order(self) -> Optional[SortOrder]:
        return SortOrder(self[0])

    @property
    def null_order(self) -> Optional[NullOrder]:
        return NullOrder(self[1])


class Field(dict):
    @staticmethod
    def of(
        field: pa.Field,
        field_id: Optional[FieldId] = None,
        is_merge_key: Optional[bool] = None,
        merge_order: Optional[MergeOrder] = None,
        is_event_time: Optional[bool] = None,
        doc: Optional[str] = None,
        past_default: Optional[Any] = None,
        future_default: Optional[Any] = None,
        consistency_type: Optional[SchemaConsistencyType] = None,
        path: Optional[NestedFieldName] = None,
        native_object: Optional[Any] = None,
    ) -> Field:
        """
        Creates a DeltaCAT field from a PyArrow base field. The DeltaCAT
        field contains a copy of the base field, but ensures that the
        PyArrow Field's metadata is also populated with optional metadata
        like documentation or metadata used within the context of a parent
        schema like field ids, merge keys, and default values.

        Args:
            field (pa.Field): Arrow base field.

            field_id (Optional[FieldId]): Unique ID of the field within its
            parent schema, or None if this field has no parent schema. If not
            given, then the field ID will be derived from the Arrow base
            field's "PARQUET:field_id" metadata key.

            is_merge_key (Optional[bool]): True if this Field is used as a merge
            key within its parent schema, False or None if it is not a merge
            key or has no parent schema. If not given, this will be derived from
            the Arrow base field's "DELTACAT:merge_key" metadata key. Merge keys
            are the default keys used to find matching records for equality
            deletes, upserts, and other equality-key-based merge operations.
            Must be a non-floating-point primitive type.

            merge_order (Optional[MergeOrder]): Merge order for this field
            within its parent schema. None if it is not used for merge order or
            has no parent schema. If not given, this will be derived from
            the Arrow base field's "DELTACAT:merge_order" metadata key. Merge
            order is used to determine the record kept amongst all records
            with matching merge keys for equality deletes, upserts, and other
            equality-key-based merge operations. Must be a primitive type.

            is_event_time (Optional[bool]): True if this Field is used to derive
            event time within its parent schema, False or None if it is not used
            or has no parent schema. If not given, this will be derived from
            the Arrow base field's "DELTACAT:event_time" metadata key. Event
            times are used to determine a stream's data completeness watermark.
            Must be an integer, float, or date type.

            doc (Optional[str]): Documentation for this field or None if this
            field has no documentation. If not given, then docs will be derived
            from the Arrow base field's "DELTACAT:doc" metadata key.

            past_default (Optional[Any]): Past default values for records
            written to the parent schema before this field was appended,
            or None if this field has no parent schema. If not given, this will
            be derived from the Arrow base field's "DELTACAT:past_default"
            metadata key. Must be coercible to the field's base arrow type.

            future_default (Optional[Any]): Future default values for records
            that omit this field in the parent schema they're written to, or
            None if this field has no parent schema. If not given, this will
            be derived from the Arrow base field's "DELTACAT:future_default"
            metadata key. Must be coercible to the field's base arrow type.

            consistency_type (Optional[SchemaConsistencyType]): Schema
            consistency type for records written to this field within the
            context of a parent schema, or None if the field has no parent
            schema. If not given, this will be derived from the Arrow base
            field's "DELTACAT:consistency_type" metadata key.

            path (Optional[NestedFieldName]): Fully qualified path of this
            field within its parent schema. Any manually specified path will
            be overwritten when this field is added to a schema.

            native_object (Optional[Any]): The native object, if any, that this
            field was originally derived from.
        Returns:
            A new DeltaCAT Field.
        """
        final_field = Field._build(
            field=field,
            field_id=Field._field_id(field) if field_id is None else field_id,
            is_merge_key=Field._is_merge_key(field)
            if is_merge_key is None
            else is_merge_key,
            merge_order=Field._merge_order(field)
            if merge_order is None
            else merge_order,
            is_event_time=Field._is_event_time(field)
            if is_event_time is None
            else is_event_time,
            doc=Field._doc(field) if doc is None else doc,
            past_default=Field._past_default(field)
            if past_default is None
            else past_default,
            future_default=Field._future_default(field)
            if future_default is None
            else future_default,
            consistency_type=Field._consistency_type(field)
            if consistency_type is None
            else consistency_type,
        )
        return Field(
            {
                "arrow": final_field,
                "path": copy.deepcopy(path),
                "nativeObject": native_object,
            }
        )

    @property
    def arrow(self) -> pa.Field:
        return self["arrow"]

    @property
    def id(self) -> Optional[FieldId]:
        return Field._field_id(self.arrow)

    @property
    def path(self) -> Optional[NestedFieldName]:
        return self.get("path")

    @property
    def is_merge_key(self) -> Optional[bool]:
        return Field._is_merge_key(self.arrow)

    @property
    def merge_order(self) -> Optional[MergeOrder]:
        return Field._merge_order(self.arrow)

    @property
    def doc(self) -> Optional[str]:
        return Field._doc(self.arrow)

    @property
    def past_default(self) -> Optional[Any]:
        return Field._past_default(self.arrow)

    @property
    def future_default(self) -> Optional[Any]:
        return Field._future_default(self.arrow)

    @property
    def consistency_type(self) -> Optional[SchemaConsistencyType]:
        return Field._consistency_type(self.arrow)

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")

    @staticmethod
    def _field_id(field: pa.Field) -> Optional[FieldId]:
        field_id = None
        if field.metadata:
            bytes_val = field.metadata.get(PARQUET_FIELD_ID_KEY_NAME)
            field_id = int(bytes_val.decode()) if bytes_val else None
        return field_id

    @staticmethod
    def _doc(field: pa.Field) -> Optional[str]:
        doc = None
        if field.metadata:
            bytes_val = field.metadata.get(FIELD_DOC_KEY_NAME)
            doc = bytes_val.decode() if bytes_val else None
        return doc

    @staticmethod
    def _is_merge_key(field: pa.Field) -> Optional[bool]:
        is_merge_key = None
        if field.metadata:
            bytes_val = field.metadata.get(FIELD_MERGE_KEY_NAME)
            is_merge_key = bool(bytes_val.decode()) if bytes_val else None
        return is_merge_key

    @staticmethod
    def _merge_order(field: pa.Field) -> Optional[MergeOrder]:
        merge_order = None
        if field.metadata:
            bytes_val = field.metadata.get(FIELD_MERGE_ORDER_KEY_NAME)
            merge_order = msgpack.loads(bytes_val) if bytes_val else None
        return merge_order

    @staticmethod
    def _is_event_time(field: pa.Field) -> Optional[bool]:
        is_event_time = None
        if field.metadata:
            bytes_val = field.metadata.get(FIELD_EVENT_TIME_KEY_NAME)
            is_event_time = bool(bytes_val.decode()) if bytes_val else None
        return is_event_time

    @staticmethod
    def _past_default(field: pa.Field) -> Optional[Any]:
        default = None
        if field.metadata:
            bytes_val = field.metadata.get(FIELD_PAST_DEFAULT_KEY_NAME)
            default = msgpack.loads(bytes_val) if bytes_val else None
        return default

    @staticmethod
    def _future_default(field: pa.Field) -> Optional[Any]:
        default = None
        if field.metadata:
            bytes_val = field.metadata.get(FIELD_FUTURE_DEFAULT_KEY_NAME)
            default = msgpack.loads(bytes_val) if bytes_val else None
        return default

    @staticmethod
    def _consistency_type(field: pa.Field) -> Optional[SchemaConsistencyType]:
        t = None
        if field.metadata:
            bytes_val = field.metadata.get(FIELD_CONSISTENCY_TYPE_KEY_NAME)
            t = SchemaConsistencyType(bytes_val.decode()) if bytes_val else None
        return t

    @staticmethod
    def _validate_merge_key(field: pa.Field):
        if not (pa.types.is_string(field.type) or pa.types.is_primitive(field.type)):
            raise ValueError(f"Merge key {field} must be a primitive type.")
        if pa.types.is_floating(field.type):
            raise ValueError(f"Merge key {field} cannot be floating point.")

    @staticmethod
    def _validate_merge_order(field: pa.Field):
        if not pa.types.is_primitive(field.type):
            raise ValueError(f"Merge order {field} must be a primitive type.")

    @staticmethod
    def _validate_event_time(field: pa.Field):
        if (
            not pa.types.is_integer(field.type)
            and not pa.types.is_floating(field.type)
            and not pa.types.is_date(field.type)
        ):
            raise ValueError(f"Event time {field} must be numeric or date type.")

    @staticmethod
    def _validate_default(
        default: Optional[Any],
        field: pa.Field,
    ) -> pa.Scalar:
        try:
            return pa.scalar(default, field.type)
        except ArrowInvalid:
            raise ValueError(
                f"Cannot treat default value `{default}` as type"
                f"`{field.type}` for field: {field}"
            )

    @staticmethod
    def _build(
        field: pa.Field,
        field_id: Optional[int],
        is_merge_key: Optional[bool],
        merge_order: Optional[MergeOrder],
        is_event_time: Optional[bool],
        doc: Optional[str],
        past_default: Optional[Any],
        future_default: Optional[Any],
        consistency_type: Optional[SchemaConsistencyType],
    ) -> pa.Field:
        meta = {}
        if is_merge_key:
            Field._validate_merge_key(field)
            meta[FIELD_MERGE_KEY_NAME] = str(is_merge_key)
        if merge_order:
            Field._validate_merge_order(field)
            meta[FIELD_MERGE_ORDER_KEY_NAME] = msgpack.dumps(merge_order)
        if is_event_time:
            Field._validate_event_time(field)
            meta[FIELD_EVENT_TIME_KEY_NAME] = str(is_event_time)
        if past_default is not None:
            Field._validate_default(past_default, field)
            meta[FIELD_PAST_DEFAULT_KEY_NAME] = msgpack.dumps(past_default)
        if future_default is not None:
            Field._validate_default(future_default, field)
            meta[FIELD_FUTURE_DEFAULT_KEY_NAME] = msgpack.dumps(future_default)
        if field_id is not None:
            meta[PARQUET_FIELD_ID_KEY_NAME] = str(field_id)
        if doc is not None:
            meta[FIELD_DOC_KEY_NAME] = doc
        if consistency_type is not None:
            meta[FIELD_CONSISTENCY_TYPE_KEY_NAME] = consistency_type.value
        return pa.field(
            name=field.name,
            type=field.type,
            nullable=field.nullable,
            metadata=meta,
        )


SingleSchema = Union[List[Field], pa.Schema]
MultiSchema = Union[Dict[SchemaName, List[Field]], Dict[SchemaName, pa.Schema]]


class Schema(dict):
    @staticmethod
    def of(
        schema: Union[SingleSchema, MultiSchema],
        schema_id: Optional[SchemaId] = None,
        native_object: Optional[Any] = None,
    ) -> Schema:
        """
        Creates a DeltaCAT schema from either one or multiple Arrow base schemas
        or lists of DeltaCAT fields. All field names across all input schemas
        must be unique (case-insensitive). If a dict of named subschemas is
        given, then this DeltaCAT schema will be backed by a unified arrow
        schema created as a union of all input schemas in the natural iteration
        order of their dictionary keys. This unified schema saves all named
        subschema field mappings in its metadata to support DeltaCAT subschema
        retrieval by name after schema creation.

        Args:
            schema (Union[SingleSchema, MultiSchema]): For a single unnamed
            schema, either an Arrow base schema or list of DeltaCAT fields.
            If an Arrow base schema is given, then a copy of the base schema
            is made with each Arrow field populated with additional metadata.
            Field IDs, merge keys, docs, and default vals will be read from
            each Arrow field's metadata if they exist. Any field missing a
            field ID will be assigned a unique field ID, with assigned field
            IDs either starting from 0 or the max field ID + 1.
            For multiple named subschemas, a dictionary of schema names to an
            arrow base schema or list of DeltaCAT fields. These schemas will
            be copied into a unified Arrow schema representing a union of all
            of their fields in their natural iteration order. Any missing
            field IDs will be autoassigned starting from 0 or the max field ID
            + 1 across the natural iteration order of all  schemas first, and
            all fields second.
            All fields across all schemas must have unique names
            (case-insensitive).

            schema_id (SchemaId): Unique ID of schema within its parent table
            version. Defaults to 0.

            native_object (Optional[Any]): The native object, if any, that this
            schema was converted from.
        Returns:
            A new DeltaCAT Schema.
        """
        # normalize the input as a unified pyarrow schema
        # if the input included multiple subschemas, then also save a mapping
        # from each subschema to its unique field names
        schema, subschema_to_field_names = Schema._to_unified_pyarrow_schema(schema)
        # discover assigned field IDs in the given pyarrow schema
        field_ids_to_fields = {}
        schema_metadata = {}
        visitor_dict = {"maxFieldId": 0}
        # find and save the schema's max field ID in the visitor dictionary
        Schema._visit_fields(
            current=schema,
            visit=Schema._find_max_field_id,
            visitor_dict=visitor_dict,
        )
        max_field_id = visitor_dict["maxFieldId"]
        visitor_dict["fieldIdsToFields"] = field_ids_to_fields
        # populate map of field IDs to DeltaCAT fields w/ IDs, docs, etc.
        Schema._visit_fields(
            current=schema,
            visit=Schema._populate_fields,
            visitor_dict=visitor_dict,
        )
        if schema.metadata:
            schema_metadata.update(schema.metadata)
        # populate merge keys
        merge_keys = [
            field.id for field in field_ids_to_fields.values() if field.is_merge_key
        ]
        # create a new pyarrow schema with field ID, doc, etc. field metadata
        pyarrow_schema = pa.schema(
            fields=[field.arrow for field in field_ids_to_fields.values()],
        )
        # map subschema field names to IDs (for faster lookup and reduced size)
        subschema_to_field_ids = {
            schema_name: [
                Field.of(pyarrow_schema.field(field_name)).id
                for field_name in field_names
            ]
            for schema_name, field_names in subschema_to_field_names.items()
        }
        # create a final pyarrow schema with populated schema metadata
        if schema_id is not None:
            schema_metadata[SCHEMA_ID_KEY_NAME] = str(schema_id)
        if schema_metadata.get(SCHEMA_ID_KEY_NAME) is None:
            schema_metadata[SCHEMA_ID_KEY_NAME] = str(0)
        schema_metadata[SUBSCHEMAS_KEY_NAME] = msgpack.dumps(subschema_to_field_ids)
        final_schema = pyarrow_schema.with_metadata(schema_metadata)
        return Schema(
            {
                "arrow": final_schema,
                "mergeKeys": merge_keys or None,
                "fieldIdsToFields": field_ids_to_fields,
                "maxFieldId": max_field_id,
                "nativeObject": native_object,
            }
        )

    @staticmethod
    def deserialize(serialized: pa.Buffer) -> Schema:
        return Schema.of(schema=pa.ipc.read_schema(serialized))

    def serialize(self) -> pa.Buffer:
        return self.arrow.serialize()

    def equivalent_to(self, other: Schema, check_metadata: bool = False):
        if other is None:
            return False
        if not isinstance(other, dict):
            return False
        if not isinstance(other, Schema):
            other = Schema(other)
        return self.arrow.equals(
            other.arrow,
            check_metadata,
        )

    def add_subschema(
        self,
        name: SchemaName,
        schema: SingleSchema,
    ) -> Schema:
        subschemas = copy.copy(self.subschemas)
        if not subschemas:  # self is SingleSchema
            subschemas = {BASE_SCHEMA_NAME: self}
        subschemas = Schema._add_subschema(name, schema, subschemas)
        return Schema.of(
            schema=subschemas,
            schema_id=self.id + 1,
        )

    def delete_subschema(self, name: SchemaName) -> Schema:
        subschemas = copy.copy(self.subschemas)
        subschemas = self._del_subschema(name, subschemas)
        if not subschemas:
            raise ValueError(f"Deleting `{name}` would leave the schema empty.")
        subschemas = {name: val.arrow for name, val in subschemas.items()}
        return Schema.of(
            schema=subschemas,
            schema_id=self.id + 1,
        )

    def replace_subschema(
        self,
        name: SchemaName,
        schema: SingleSchema,
    ) -> Schema:
        subschemas = copy.copy(self.subschemas)
        subschemas = Schema._del_subschema(name, subschemas)
        subschemas = Schema._add_subschema(name, schema, subschemas)
        return Schema.of(
            schema=subschemas,
            schema_id=self.id + 1,
        )

    def field_id(self, name: Union[FieldName, NestedFieldName]) -> FieldId:
        return Schema._field_name_to_field_id(self.arrow, name)

    def field_name(self, field_id: FieldId) -> Union[FieldName, NestedFieldName]:
        field = self.field_ids_to_fields[field_id]
        if len(field.path) == 1:
            return field.arrow.name
        return field.path

    def field(self, field_locator: FieldLocator) -> Field:
        field_id = (
            field_locator
            if isinstance(field_locator, FieldId)
            else self.field_id(field_locator)
        )
        return self.field_ids_to_fields[field_id]

    @property
    def fields(self) -> List[Field]:
        field_ids_to_fields = self.field_ids_to_fields
        return list(field_ids_to_fields.values())

    @property
    def merge_keys(self) -> Optional[List[FieldId]]:
        return self.get("mergeKeys")

    @property
    def field_ids_to_fields(self) -> Dict[FieldId, Field]:
        return self.get("fieldIdsToFields")

    @property
    def arrow(self) -> pa.Schema:
        return self["arrow"]

    @property
    def max_field_id(self) -> FieldId:
        return self["maxFieldId"]

    @property
    def id(self) -> SchemaId:
        return Schema._schema_id(self.arrow)

    @property
    def subschema(self, name: SchemaName) -> Optional[Schema]:
        subschemas = self.subschemas
        return subschemas.get(name) if subschemas else None

    @property
    def subschemas(self) -> Dict[SchemaName, Schema]:
        # return cached subschemas first if they exist
        subschemas = self.get("subschemas")
        if not subschemas:
            # retrieve any defined subschemas
            subschemas_to_field_ids = self.subschemas_to_field_ids
            # rebuild and return the subschema cache
            if subschemas_to_field_ids:
                subschemas = {
                    schema_name: Schema.of(
                        schema=pa.schema(
                            [self.field(field_id).arrow for field_id in field_ids]
                        ),
                        schema_id=self.id,
                        native_object=self.native_object,
                    )
                    for schema_name, field_ids in subschemas_to_field_ids.items()
                }
                self["subschemas"] = subschemas
        return subschemas or {}

    @property
    def subschema_field_ids(self, name: SchemaName) -> Optional[List[FieldId]]:
        return self.subschemas_to_field_ids.get(name)

    @property
    def subschemas_to_field_ids(self) -> Dict[SchemaName, List[FieldId]]:
        return Schema._subschemas(self.arrow)

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")

    @staticmethod
    def _schema_id(schema: pa.Schema) -> SchemaId:
        schema_id = None
        if schema.metadata:
            bytes_val = schema.metadata.get(SCHEMA_ID_KEY_NAME)
            schema_id = int(bytes_val.decode()) if bytes_val else None
        return schema_id

    @staticmethod
    def _subschemas(
        schema: pa.Schema,
    ) -> Dict[SchemaName, List[FieldId]]:
        subschemas = None
        if schema.metadata:
            bytes_val = schema.metadata.get(SUBSCHEMAS_KEY_NAME)
            subschemas = msgpack.loads(bytes_val) if bytes_val else None
        return subschemas

    @staticmethod
    def _field_name_to_field_id(
        schema: pa.Schema,
        name: Union[FieldName, NestedFieldName],
    ) -> FieldId:
        if isinstance(name, str):
            return Field.of(schema.field(name)).id
        if isinstance(name, List):
            if not len(name):
                raise ValueError(f"Nested field name `{name}` is empty.")
            field = schema
            for part in name:
                field = field[part]
            return Field.of(field).id
        raise ValueError(f"Unknown field name type: {type(name)}")

    @staticmethod
    def _visit_fields(
        current: Union[pa.Schema, pa.Field],
        visit: Callable,
        path: NestedFieldName = [],
        *args,
        **kwargs,
    ) -> None:
        """
        Recursively visit all fields in a PyArrow schema, including nested
        fields.

        Args:
            current (pa.Schema or pa.Field): The schema or field to visit.
            visit (callable): A function that visits the current field.
            path (NestedFieldName): The current path to the field.
            *args: Additional args to pass to the visit function.
            **kwargs: Additional keyword args to pass to the visit function.
        Returns:
            None
        """
        if isinstance(current, pa.Schema):
            for field in current:
                Schema._visit_fields(
                    field,
                    visit,
                    path,
                    *args,
                    **kwargs,
                )
        elif isinstance(current, pa.Field):
            path.append(current.name)
            visit(current, path, *args, **kwargs)
            if pa.types.is_nested(current.type):
                if isinstance(current, pa.StructType):
                    for field in current:
                        Schema._visit_fields(
                            field,
                            visit,
                            path,
                            *args,
                            **kwargs,
                        )
                elif isinstance(current, pa.ListType):
                    Schema._visit_fields(
                        current.value_field,
                        visit,
                        path,
                        *args,
                        **kwargs,
                    )
                elif isinstance(current, pa.MapType):
                    Schema._visit_fields(
                        current.key_field,
                        visit,
                        path,
                        *args,
                        **kwargs,
                    )
                    Schema._visit_fields(
                        current.item_field,
                        visit,
                        path,
                        *args,
                        **kwargs,
                    )
            path.pop()
        else:
            raise ValueError(f"Unexpected Schema Field Type: {type(current)}")

    @staticmethod
    def _find_max_field_id(
        field: pa.Field,
        path: NestedFieldName,
        visitor_dict: Dict[str, Any],
    ) -> None:
        max_field_id = max(
            visitor_dict.get("maxFieldId", 0),
            Field.of(field).id or 0,
        )
        visitor_dict["maxFieldId"] = max_field_id

    @staticmethod
    def _populate_fields(
        field: pa.Field,
        path: NestedFieldName,
        visitor_dict: Dict[str, Any],
    ) -> None:
        field_ids_to_fields = visitor_dict["fieldIdsToFields"]
        max_field_id = (
            visitor_dict["maxFieldId"] + len(field_ids_to_fields)
        ) % MAX_FIELD_ID_EXCLUSIVE
        dc_field = Field.of(field)
        if dc_field is not None and dc_field.id is not None:
            field_id = dc_field.id
        else:
            field_id = max_field_id

        if (dupe := field_ids_to_fields.get(field_id)) is not None:
            raise ValueError(
                f"Duplicate field id {field_id} for field: {field} "
                f"Already assigned to field: {dupe}"
            )
        field = Field.of(
            field=field,
            field_id=field_id,
            path=path,
        )
        field_ids_to_fields[field_id] = field

    @staticmethod
    def _get_lower_case_field_names(
        schema: SingleSchema,
    ) -> List[str]:
        if isinstance(schema, pa.Schema):
            return [name.lower() for name in schema.names]
        elif isinstance(schema, List):  # List[Field]
            names = [f.arrow.name.lower() for f in schema if isinstance(f, Field)]
            if len(names) == len(schema):
                return names  # all items in list are valid Field objects
        raise ValueError(f"Unsupported schema argument: {schema}")

    @staticmethod
    def _validate_schema_name(name: str) -> None:
        if not name:
            raise ValueError(f"Schema name cannot be empty.")
        if len(name) > BYTES_PER_KIBIBYTE:
            raise ValueError(
                f"Invalid schema name `{name}`. Schema names "
                f"cannot be greater than {BYTES_PER_KIBIBYTE} "
                f"characters."
            )

    @staticmethod
    def _validate_field_names(
        schema: Union[SingleSchema, MultiSchema],
    ) -> None:
        all_names = []
        if isinstance(schema, dict):  # MultiSchema
            for schema_name, val in schema.items():
                Schema._validate_schema_name(schema_name)
                all_names.extend(Schema._get_lower_case_field_names(val))
        else:  # SingleSchema
            all_names.extend(Schema._get_lower_case_field_names(schema))
        if not all_names:
            raise ValueError(f"Schema must contain at least one field.")
        name_set = set()
        dupes = []
        for name in all_names:
            dupes.append(name) if name in name_set else name_set.add(name)
        if dupes:
            raise ValueError(
                f"Expected all schema fields to have unique names "
                f"(case-insensitive), but found the following duplicates: "
                f"{dupes}"
            )

    @staticmethod
    def _to_pyarrow_schema(schema: SingleSchema) -> pa.Schema:
        if isinstance(schema, pa.Schema):
            return schema
        elif isinstance(schema, List):  # List[Field]
            return pa.schema(fields=[field.arrow for field in schema])
        else:
            raise ValueError(f"Unsupported schema base type: {schema}")

    @staticmethod
    def _to_unified_pyarrow_schema(
        schema: Union[SingleSchema, MultiSchema],
    ) -> Tuple[pa.Schema, Dict[SchemaName, List[FieldName]]]:
        # first, ensure all field names are valid and contain no duplicates
        Schema._validate_field_names(schema)
        # now union all schemas into a single schema
        subschema_to_field_names = {}
        if isinstance(schema, dict):  # MultiSchema
            all_schemas = []
            for schema_name, schema_val in schema.items():
                pyarow_schema = Schema._to_pyarrow_schema(schema_val)
                all_schemas.append(pyarow_schema)
                subschema_to_field_names[schema_name] = [
                    field.name for field in pyarow_schema
                ]
            return pa.unify_schemas(all_schemas), subschema_to_field_names
        return Schema._to_pyarrow_schema(schema), {}  # SingleSchema

    @staticmethod
    def _del_subschema(
        name: SchemaName,
        subschemas: Dict[SchemaName, Schema],
    ) -> Dict[SchemaName, Schema]:
        deleted_subschema = subschemas.pop(name, None)
        if deleted_subschema is None:
            raise ValueError(f"Subschema `{name}` does not exist.")
        return subschemas

    @staticmethod
    def _add_subschema(
        name: SchemaName,
        schema: SingleSchema,
        subschemas: Dict[SchemaName, Schema],
    ) -> Dict[SchemaName, Schema]:
        Schema._validate_schema_name(name)
        if name == BASE_SCHEMA_NAME:
            raise ValueError(
                f"Cannot add subschema with reserved name: {BASE_SCHEMA_NAME}"
            )
        if name in subschemas:
            raise ValueError(f"Subschema `{name}` already exists.")
        for key, val in subschemas.items():
            subschemas[key] = val.arrow
        subschemas[name] = schema
        return subschemas


class SchemaList(List[Schema]):
    @staticmethod
    def of(items: List[Schema]) -> SchemaList:
        typed_items = SchemaList()
        for item in items:
            if item is not None and not isinstance(item, Schema):
                item = Schema(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, Schema):
            self[item] = val = Schema(val)
        return val



from typing import Iterator

from daft import Schema
from daft.daft import (
    StorageConfig,
    PartitionField,
    Pushdowns,
    ScanTask,
    FileFormatConfig,
    ParquetSourceConfig,
)
from daft.io.scan import ScanOperator

from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.daft.model import DaftPartitionKeyMapper


class DeltaCatScanOperator(ScanOperator):
    def __init__(self, table: TableDefinition, storage_config: StorageConfig) -> None:
        super().__init__()
        self.table = table
        self._schema = self._infer_schema()
        self.partition_keys = self._infer_partition_keys()
        self.storage_config = storage_config

    def schema(self) -> Schema:
        return self._schema

    def name(self) -> str:
        return "DeltaCatScanOperator"

    def display_name(self) -> str:
        return f"DeltaCATScanOperator({self.table.table.namespace}.{self.table.table.table_name})"

    def partitioning_keys(self) -> list[PartitionField]:
        return self.partition_keys

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self._schema}",
            f"Partitioning keys = {self.partitioning_keys}",
            f"Storage config = {self.storage_config}",
        ]

    def to_scan_tasks(self, pushdowns: Pushdowns) -> Iterator[ScanTask]:
        # TODO: implement pushdown predicate on DeltaCAT
        dc_scan_plan = self.table.create_scan_plan()
        scan_tasks = []
        file_format_config = FileFormatConfig.from_parquet_config(
            # maybe this: ParquetSourceConfig(field_id_mapping=self._field_id_mapping)
            ParquetSourceConfig()
        )
        for dc_scan_task in dc_scan_plan.scan_tasks:
            for data_file in dc_scan_task.data_files():
                st = ScanTask.catalog_scan_task(
                    file=data_file.file_path,
                    file_format=file_format_config,
                    schema=self._schema._schema,
                    storage_config=self.storage_config,
                    pushdowns=pushdowns,
                )
                scan_tasks.append(st)
        return iter(scan_tasks)

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return True

    def _infer_schema(self) -> Schema:

        if not (
            self.table and self.table.table_version and self.table.table_version.schema
        ):
            raise RuntimeError(
                f"Failed to infer schema for DeltaCAT Table "
                f"{self.table.table.namespace}.{self.table.table.table_name}"
            )

        return Schema.from_pyarrow_schema(self.table.table_version.schema.arrow)

    def _infer_partition_keys(self) -> list[PartitionField]:
        if not (
            self.table
            and self.table.table_version
            and self.table.table_version.partition_scheme
            and self.table.table_version.schema
        ):
            raise RuntimeError(
                f"Failed to infer partition keys for DeltaCAT Table "
                f"{self.table.table.namespace}.{self.table.table.table_name}"
            )

        schema = self.table.table_version.schema
        partition_keys = self.table.table_version.partition_scheme.keys
        if not partition_keys:
            return []

        partition_fields = []
        for key in partition_keys:
            field = DaftPartitionKeyMapper.unmap(key, schema)
            # Assert that the returned value is not None.
            assert field is not None, f"Unmapping failed for key {key}"
            partition_fields.append(field)

        return partition_fields
